#ifndef OPENTISSUE_UTILITY_IMAGE_ANALYSIS_IMAGE_ANALYSIS_COMPUTE_FEATURE_SPACE_H
#define OPENTISSUE_UTILITY_IMAGE_ANALYSIS_IMAGE_ANALYSIS_COMPUTE_FEATURE_SPACE_H
//
// OpenTissue Template Library
// - A generic toolbox for physics-based modeling and simulation.
// Copyright (C) 2008 Department of Computer Science, University of Copenhagen.
//
// OTTL is licensed under zlib: http://opensource.org/licenses/zlib-license.php
//
#include <OpenTissue/configuration.h>

#include <OpenTissue/core/containers/grid/grid.h>
#include <OpenTissue/core/containers/grid/util/grid_gradient_magnitude.h>
#include <OpenTissue/core/containers/grid/util/grid_second_derivative.h>
#include <OpenTissue/core/math/math_is_finite.h>


namespace OpenTissue
{
  namespace image_analysis
  {

    /**
    * Compute Histogram Space Grid.
    *
    * @param U                       The image to compute a histogram of.
    * @param feature_grid             Each voxel contains a triplet correspond to intensity, gradient magnitude
    *                                and second derivatives of the corresponding voxel in U.
    */
    template<
        typename image_type
      , typename feature_grid_type // = OpenTissue::grid::Grid< Vector3<double> >
    >
    void compute_feature_space(
        image_type const & U
      , feature_grid_type & feature_grid
    )
    {
      typedef typename feature_grid_type::math_types         math_types;
      typedef OpenTissue::grid::Grid<double,math_types>      real_grid_type;
      typedef typename real_grid_type::index_iterator        image_iterator;
      typedef typename feature_grid_type::value_type         vector3_type;
      typedef typename feature_grid_type::index_iterator     feature_iterator;

      //--- Convert U into reals to ease problems with data type conversions
      //--- This is a temporary storage, so no big harm is done!!!
      real_grid_type real_U;
      real_U.create(U.min_coord(),U.max_coord(),U.I(),U.J(),U.K());
      std::copy(U.begin(),U.end(),real_U.begin());

      //--- Allocate memory for feature vectors.
      feature_grid.create(U.min_coord(),U.max_coord(),U.I(),U.J(),U.K());

      feature_iterator fbegin = feature_grid.begin();
      feature_iterator fend   = feature_grid.end();
      feature_iterator f      = fbegin;

      image_iterator ubegin   = real_U.begin();
      image_iterator uend     = real_U.end();
      image_iterator u        = ubegin;

      for(; u!=uend; ++u,++f)
      {
        (*f)(0) = *u;
        (*f)(1) = OpenTissue::grid::gradient_magnitude(u);
        (*f)(2) = OpenTissue::grid::second_derivative(u);

        assert( is_finite( (*f)(0) ) || !"compute_feature_space(): NaN");
        assert( is_finite( (*f)(1) ) || !"compute_feature_space(): NaN" );
        assert( is_finite( (*f)(2) ) || !"compute_feature_space(): NaN" );
      }
    }

  }// end namespace image_analysis

} // namespace OpenTissue

//  OPENTISSUE_UTILITY_IMAGE_ANALYSIS_IMAGE_ANALYSIS_COMPUTE_FEATURE_SPACE_H
#endif
