#ifndef OPENTISSUE_UTILITY_IMAGE_ANALYSIS_IMAGE_ANALYSIS_COMPUTE_FEATURE_SPACE_HISTOGRAM_H
#define OPENTISSUE_UTILITY_IMAGE_ANALYSIS_IMAGE_ANALYSIS_COMPUTE_FEATURE_SPACE_HISTOGRAM_H
//
// OpenTissue Template Library
// - A generic toolbox for physics-based modeling and simulation.
// Copyright (C) 2008 Department of Computer Science, University of Copenhagen.
//
// OTTL is licensed under zlib: http://opensource.org/licenses/zlib-license.php
//
#include <OpenTissue/configuration.h>

#include <OpenTissue/core/math/math_is_finite.h> 

namespace OpenTissue
{
  namespace image_analysis
  {
    /**
     * Create Feature Space Histogram.
     *
     * @param feature_grid    The feature space grid.
     * @param I              Number of intensity buckets in histogram.
     * @param J              Number of gradient magnitude buckets in histogram.
     * @param K              Number of second derivative buckets in histogram.
     * @param histogram      Upon return holds a histogram of the feature space.
     */
    template<
        typename feature_grid_type
      , typename histogram_type  // = OpenTissue::grid::Grid< size_t>
    >
    void compute_feature_space_histogram(
        feature_grid_type const & feature_grid
      , size_t const & I
      , size_t const & J
      , size_t const & K
      , histogram_type & histogram
    )
    {
      using std::min;
      using std::max;
      using std::floor;

      assert(I>0 || !"compute_feature_space_histogram(): I must be positive");
      assert(J>0 || !"compute_feature_space_histogram(): J must be positive");
      assert(K>0 || !"compute_feature_space_histogram(): K must be positive");

      typedef typename feature_grid_type::const_index_iterator       feature_iterator;
      typedef typename feature_grid_type::value_type                 vector3_type;
      typedef typename vector3_type::value_type                     real_type;
      typedef typename histogram_type::index_iterator               histogram_iterator;
      typedef typename histogram_type::index_type                   index_type;

      real_type min_x = feature_grid.size();
      real_type min_y = feature_grid.size();
      real_type min_z = feature_grid.size();
      real_type max_x = -min_x;
      real_type max_y = -min_y;
      real_type max_z = -min_z;

      feature_iterator fbegin   = feature_grid.begin();
      feature_iterator fend     = feature_grid.end();
      feature_iterator f;
      for(f=fbegin; f!=fend;++f)
      {
        min_x = min( (*f)(0), min_x);
        min_y = min( (*f)(1), min_y);
        min_z = min( (*f)(2), min_z);
        max_x = max( (*f)(0), max_x);
        max_y = max( (*f)(1), max_y);
        max_z = max( (*f)(2), max_z);
      }

      assert( is_finite( min_x ) || !"compute_feature_space_histogram(): NaN" );
      assert( is_finite( min_y ) || !"compute_feature_space_histogram(): NaN" );
      assert( is_finite( min_z ) || !"compute_feature_space_histogram(): NaN" );
      assert( is_finite( max_x ) || !"compute_feature_space_histogram(): NaN" );
      assert( is_finite( max_y ) || !"compute_feature_space_histogram(): NaN" );
      assert( is_finite( max_z ) || !"compute_feature_space_histogram(): NaN" );

      histogram.create(vector3_type(min_x,min_y,min_z),vector3_type(max_x,max_y,max_z),I,J,K);
      real_type dx = histogram.dx();
      real_type dy = histogram.dy();
      real_type dz = histogram.dz();

      histogram_iterator hbegin = histogram.begin();
      histogram_iterator hend   = histogram.end();
      histogram_iterator h;
      for(h=hbegin; h!=hend;++h)
        *h = 0;

      for(f = fbegin; f!=fend;++f)
      {
        index_type i = static_cast<index_type>(floor(((*f)(0) - min_x) / dx));
        index_type j = static_cast<index_type>(floor(((*f)(1) - min_y) / dy));
        index_type k = static_cast<index_type>(floor(((*f)(2) - min_z) / dz));

        if(i==I) --i;
        if(j==J) --j;
        if(k==K) --k;

        assert(i>=0 || !"compute_feature_space_histogram(): i must be non-negative" );
        assert(j>=0 || !"compute_feature_space_histogram(): j must be non-negative" );
        assert(k>=0 || !"compute_feature_space_histogram(): k must be non-negative" );
        assert(i<I  || !"compute_feature_space_histogram(): i must be less than I" );
        assert(j<J  || !"compute_feature_space_histogram(): j must be less than J" );
        assert(k<K  || !"compute_feature_space_histogram(): k must be less than K" );

        ++histogram(i,j,k);
      }
    }

  }// end namespace image_analysis

} // namespace OpenTissue

//  OPENTISSUE_UTILITY_IMAGE_ANALYSIS_IMAGE_ANALYSIS_COMPUTE_FEATURE_SPACE_HISTOGRAM_H
#endif
