#ifndef OPENTISSUE_UTILITY_IMAGE_ANALYSIS_IMAGE_ANALYSIS_COMPUTE_FEATURE_SPACE_WHITENING_H
#define OPENTISSUE_UTILITY_IMAGE_ANALYSIS_IMAGE_ANALYSIS_COMPUTE_FEATURE_SPACE_WHITENING_H
//
// OpenTissue Template Library
// - A generic toolbox for physics-based modeling and simulation.
// Copyright (C) 2008 Department of Computer Science, University of Copenhagen.
//
// OTTL is licensed under zlib: http://opensource.org/licenses/zlib-license.php
//
#include <OpenTissue/configuration.h>

#include <OpenTissue/utility/image_analysis/feature_space_pca.h>
#include <OpenTissue/core/math/math_eigen_system_decomposition.h>
#include <OpenTissue/core/math/math_is_finite.h>

namespace OpenTissue
{
  namespace image_analysis
  {
    /**
     * Feature Space Whitening.
     * This function rotates and scales feature space according to a PCA analysis. 
     *
     *
     * @param feature_grid    The feature space grid.
     * @param R              Upon return this argument holds the rotational warp applied to the feature space.
     * @param d              Upon return this argument holds a 3D vector containing the scaling along the x,y and z axes.
     */
    template< typename feature_grid_type, typename matrix3x3_type, typename vector3_type >
    void compute_feature_space_whitening(feature_grid_type & feature_grid, matrix3x3_type & R, vector3_type & d  )
    {
      using std::sqrt;

      typedef typename feature_grid_type::index_iterator     feature_iterator;

      matrix3x3_type C;
      vector3_type mean;
      OpenTissue::image_analysis::feature_space_pca(feature_grid,mean,C);

      OpenTissue::math::eigen(C,R,d);
      //vector3_type e0 = vector3_type( R(0,0), R(1,0), R(2,0) );
      //vector3_type e1 = vector3_type( R(0,1), R(1,1), R(2,1) );
      //vector3_type e2 = vector3_type( R(0,2), R(1,2), R(2,2) );
      vector3_type inv_d2;
      inv_d2(0) = 1.0 / sqrt(d(0));
      inv_d2(1) = 1.0 / sqrt(d(1));
      inv_d2(2) = 1.0 / sqrt(d(2));
      feature_iterator fbegin = feature_grid.begin();
      feature_iterator fend   = feature_grid.end();
      feature_iterator f      = fbegin;
      for(; f!=fend;++f)
      {
        (*f) = R*((*f)-mean);
        (*f)(0) *= inv_d2(0);
        (*f)(1) *= inv_d2(1);
        (*f)(2) *= inv_d2(2);
        assert( is_finite( (*f)(0) ) || !"compute_feature_space_whitening(): NaN ");
        assert( is_finite( (*f)(1) ) || !"compute_feature_space_whitening(): NaN " );
        assert( is_finite( (*f)(2) ) || !"compute_feature_space_whitening(): NaN " );
      }
    }

  }// end namespace image_analysis

} // namespace OpenTissue

//  OPENTISSUE_UTILITY_IMAGE_ANALYSIS_IMAGE_ANALYSIS_COMPUTE_FEATURE_SPACE_WHITENING_H
#endif
