#ifndef OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_COLOR_TABLE_H
#define OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_COLOR_TABLE_H
//
// OpenTissue Template Library
// - A generic toolbox for physics-based modeling and simulation.
// Copyright (C) 2008 Department of Computer Science, University of Copenhagen.
//
// OTTL is licensed under zlib: http://opensource.org/licenses/zlib-license.php
//
#include <OpenTissue/configuration.h>

#include <OpenTissue/gpu/image/image.h>
#include <OpenTissue/gpu/image/image_make_noise.h>
#include <OpenTissue/gpu/texture/texture_texture2D.h>

#include <OpenTissue/core/math/math_constants.h>


namespace OpenTissue
{
  namespace volume_rendering
  {

    /**
    * Convinience Types.
    * These types make it easy to create color tables in applications.
    */
    typedef OpenTissue::image::Image<unsigned char>                 color_table_type;
    typedef boost::shared_ptr<color_table_type>  color_table_pointer;

    /**
    * Histogram Based Color Table.
    * This function analyses a histogram and assign colors to
    * a color table based on the features detected in the
    * histogram.
    *
    */
    template< typename histogram_type, typename image_type >
    void compute_histogram_color_table( unsigned int entries, histogram_type const & histogram, image_type & color_table )
    {
      using std::fabs;

      typedef typename image_type::value_type  value_type;
      typedef          float                   real_type;

      unsigned int N = histogram.size();
      std::vector<real_type> H,D;
      H.resize(N);
      D.resize(N);
      for(unsigned int idx=0;idx<N;++idx)
        H[idx] = histogram[idx];

      //--- Compute first derivatives
      D[0]   = H[1] - H[0];
      D[N-1] = H[N-2] - H[N-1];
      for(unsigned int i=1;i<N-1;++i)
        D[i] =  (H[i+1] - H[i-1])*.5;

      //--- Detect regions of same slope
      assert(entries>0 || !"");
      std::vector<unsigned int>  region_lut;
      region_lut.resize(entries);
      real_type  nominator = static_cast<real_type>(entries);

      real_type old_slope = D[0];
      unsigned int region = 0;
      region_lut[0] = region;
      for(unsigned int idx=1;idx<entries;++idx)
      {
        unsigned int i      =  static_cast<unsigned int>(  ( idx /nominator) *N );
        real_type cur_slope = D[i];
        //if(fabs(cur_slope - old_slope)>1.0)
        if(cur_slope*old_slope<0.0)
        {
          ++region;
        }
        region_lut[idx] = region;
        old_slope = cur_slope;
      }
      //--- create color table
      assert(entries>0 || !"");
      color_table.create(entries, 1, 4);
      value_type  high      = math::detail::highest<value_type>();
      value_type  low       = math::detail::lowest<value_type>();
      for(unsigned int i=0;i<entries;++i)
      {
        real_type  t       = (1.0*region_lut[i]) / (1.0*region);
        value_type I       =  static_cast<value_type>( t * (high-low)  + low );      
        //if((region_lut[i]%7)==0)
        //{
        //  color_table(i,0,0) = low;      color_table(i,0,1) = low;      color_table(i,0,2) = I;
        //}
        //else if((region_lut[i]%7)==1)
        //{
        //  color_table(i,0,0) = low;      color_table(i,0,1) = I;      color_table(i,0,2) = low;
        //}
        //else if((region_lut[i]%7)==2)
        //{
        //  color_table(i,0,0) = I;      color_table(i,0,1) = low;      color_table(i,0,2) = low;
        //}
        //else if((region_lut[i]%7)==3)
        //{
        //  color_table(i,0,0) = low;      color_table(i,0,1) = I;      color_table(i,0,2) = I;
        //}
        //else if((region_lut[i]%7)==4)
        //{
        //  color_table(i,0,0) = I;      color_table(i,0,1) = low;      color_table(i,0,2) = I;
        //}
        //else if((region_lut[i]%7)==5)
        //{
        //  color_table(i,0,0) = I;      color_table(i,0,1) = I;      color_table(i,0,2) = low;
        //}
        //else if((region_lut[i]%7)==6)
        //{
        //  color_table(i,0,0) = I;      color_table(i,0,1) = I;      color_table(i,0,2) = I;
        //}
        color_table(i,0,0) = I;      color_table(i,0,1) = I;      color_table(i,0,2) = I;
        color_table(i,0,3) = high; 
      }
    }

    /**
    * Linear Color Mapping.
    *
    * @param entries       The number of entries on the color table.
    * @param color_table   Upon return this argument holds the newly
    *                      generated color table as an 1 row image with
    *                      entries pixels in it.
    * @param reverse       If set to true (default false) low image intensities
    *                      are mapped to high colors.
    * @param transparency  If set to trur the alpha channel in the color table
    *                      is also lineary scaled. Default is false.
    *
    */
    template<typename image_type >
    void compute_linear_color_table( unsigned int entries, image_type & color_table, bool reverse = false, bool transparency = false)
    {
      typedef typename image_type::value_type  value_type;
      typedef float                            real_type;
      assert(entries>0 || !"");
      color_table.create(entries, 1, 4);

      value_type  high = math::detail::highest<value_type>();
      value_type  low = math::detail::lowest<value_type>();

      real_type nominator = static_cast<real_type>(entries);
      for(unsigned int i=0;i<entries;++i)
      {
        real_type t =  (i /nominator) ;
        value_type tmp =  static_cast<value_type>( t * (high-low)  + low );
        value_type I  = reverse? high - tmp: tmp;
        color_table(i,0,0) = I;
        color_table(i,0,1) = I;
        color_table(i,0,2) = I;
        color_table(i,0,3) = transparency?I:high; 
      }
    }



    /**
    * Smooth Noise Color Mapping.
    *
    * @param entries       The number of entries on the color table.
    * @param color_table   Upon return this argument holds the newly
    *                      generated color table as an 1 row image with
    *                      entries pixels in it.
    *
    */
    template<typename image_type >
    void compute_noise_color_table( unsigned int entries, image_type & color_table)
    {
      typedef typename image_type::value_type  value_type;
      typedef float                            real_type;
      assert(entries>0 || !"");
      color_table.create(entries, 1, 4);
      make_noise(color_table);
    }


  } // namespace volume_rendering
} // namespace OpenTissue

// OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_COLOR_TABLE_H
#endif
