#ifndef OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_PREINTEGRATED_COLOR_TABLE_H
#define OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_PREINTEGRATED_COLOR_TABLE_H
//
// OpenTissue Template Library
// - A generic toolbox for physics-based modeling and simulation.
// Copyright (C) 2008 Department of Computer Science, University of Copenhagen.
//
// OTTL is licensed under zlib: http://opensource.org/licenses/zlib-license.php
//
#include <OpenTissue/configuration.h>

#include <OpenTissue/gpu/texture/texture_texture2D.h>

namespace OpenTissue
{
  namespace volume_rendering
  {

    /**
    * Compute Preintegrated Color Table.
    *
    * @param color_table    The color table that should be integrated.
    *
    * @return               A pointer to a preintegrated color table texture object.
    */
    template<typename image_type>
    OpenTissue::texture::texture2D_pointer compute_1D_preintegration_color_table( image_type const & color_table )
    {
      typedef typename image_type::value_type  value_type;
      typedef float                            real_type;

      unsigned int entries  = color_table.width();
      unsigned int rows     = color_table.height();
      unsigned int channels = color_table.channels();

      assert(rows==1     || !"compute_1D_preintegration_color_table(): Color table image must have exactly one row");
      assert(entries>0   || !"compute_1D_preintegration_color_table(): Color table image must positive number of entries");
      assert(channels==4 || !"compute_1D_preintegration_color_table(): Color table image must have four color components");

      image_type preintegration; 
      preintegration.create(entries, 1, channels);

      real_type  tauc = 0;
      real_type  r    = 0;
      real_type  g    = 0;
      real_type  b    = 0;
      real_type  a    = 0;

      preintegration(0,0,0) = r;
      preintegration(0,0,1) = g;
      preintegration(0,0,2) = b;
      preintegration(0,0,3) = a;
      for ( unsigned int i = 1; i < entries; ++i )
      {
        real_type alpha    = color_table(  i,0,3);
        real_type alpha_m1 = color_table(i-1,0,3);
        real_type red      = color_table(  i,0,0);
        real_type red_m1   = color_table(i-1,0,0);
        real_type green    = color_table(  i,0,1);
        real_type green_m1 = color_table(i-1,0,1);
        real_type blue     = color_table(  i,0,2);
        real_type blue_m1  = color_table(i-1,0,2);

        tauc =  ( alpha_m1 + alpha ) / 2.0;
        r = r + ( red_m1   + red   ) / 2.0 * tauc;
        g = g + ( green_m1 + green ) / 2.0 * tauc;
        b = b + ( blue_m1  + blue  ) / 2.0 * tauc;
        a = a + tauc;

        preintegration(i,0,0) = r;
        preintegration(i,0,1) = g;
        preintegration(i,0,2) = b;
        preintegration(i,0,3) = a;
      }
      return preintegration.create_texture(GL_RGBA);
    }

  } // namespace volume_rendering
} // namespace OpenTissue

// OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_PREINTEGRATED_COLOR_TABLE_H
#endif
