#ifndef OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_RENDER_H
#define OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_RENDER_H
//
// OpenTissue Template Library
// - A generic toolbox for physics-based modeling and simulation.
// Copyright (C) 2008 Department of Computer Science, University of Copenhagen.
//
// OTTL is licensed under zlib: http://opensource.org/licenses/zlib-license.php
//
#include <OpenTissue/configuration.h>

#include <OpenTissue/utility/gl/gl_matrix_util.h>
#include <OpenTissue/gpu/texture/texture_create_texture3D.h>
#include <OpenTissue/utility/volume_rendering/volume_rendering_texture3d_tile.h>

#include <list>
#include <cassert>

namespace OpenTissue
{
  namespace volume_rendering
  {

    /**
    * Volume Render.
    * This render class creates one big texture tile, enclosing the
    * volume. Thus if the volume is non-power of two the smallest
    * enclosing power of two cubic texture is computed and the volume
    * is loaded into a sub-region of this texture.
    */
    template< 
      typename math_types
      , template <typename> class volume_shader_class_   ///< A volume shader class inherited from the concept VolumeShader (using CRTP)
    >
    class Render
    {
    public:

      typedef typename math_types::real_type         real_type;
      typedef typename math_types::vector3_type      vector3_type;
      typedef Texture3DTile<math_types>              tile_type;
      typedef volume_shader_class_<math_types>       volume_shader_type;

    protected:

      real_type                 m_image_min_x;            ///< The minimum x-coordinate of AABB enclosing all voxel data. Used for settting up texture tiles.
      real_type                 m_image_min_y;            ///< The minimum y-coordinate of AABB enclosing all voxel data. Used for settting up texture tiles.
      real_type                 m_image_min_z;            ///< The minimum z-coordinate of AABB enclosing all voxel data. Used for settting up texture tiles.
      real_type                 m_image_width;            ///< The width of the AABB enclosing all the voxel data. Used for settting up texture tiles.
      real_type                 m_image_height;           ///< The height of the AABB enclosing all the voxel data. Used for settting up texture tiles.
      real_type                 m_image_depth;            ///< The depth of the AABB enclosing all the voxel data. Used for settting up texture tiles.
      tile_type                 m_tile;                   ///< The texture tile.
      volume_shader_type        m_shader;                 ///< The shader.    
      vector3_type              m_vertices[8];            ///< Model space vertices of the AABBB enclosing all the voxel data.

    public:

      Render()
        : m_image_min_x()
        , m_image_min_y()
        , m_image_min_z()
        , m_image_width()
        , m_image_height()
        , m_image_depth()
        , m_tile()
        , m_shader()
      { }

    public:

      /**
      * Initialization.
      * Should be invoked before attempting to visualize the volume.
      * If volume changes over time, there is no need to re-invoke the
      * method, just simply update the corresponding volume texture.
      *
      * @param volume        The volume (voxel data) that should be visualized.
      */
      template< typename grid_type >
      void init( grid_type & volume )
      {
        using std::max;

        typedef typename grid_type::value_type  T;

        real_type dx   = volume.dx();
        real_type dy   = volume.dy();
        real_type dz   = volume.dz();
        unsigned int I = volume.I();
        unsigned int J = volume.J();
        unsigned int K = volume.K();

        m_image_min_x         = volume.min_coord(0) - dx*.5;
        m_image_min_y         = volume.min_coord(1) - dy*.5;
        m_image_min_z         = volume.min_coord(2) - dz*.5;
        real_type image_max_x = volume.max_coord(0) + dx*.5;
        real_type image_max_y = volume.max_coord(1) + dy*.5;
        real_type image_max_z = volume.max_coord(2) + dz*.5;
        m_image_width         = image_max_x - m_image_min_x;
        m_image_height        = image_max_y - m_image_min_y;
        m_image_depth         = image_max_z - m_image_min_z;

        m_vertices[ 0 ] = vector3_type( m_image_min_x , m_image_min_y,  m_image_min_z );
        m_vertices[ 1 ] = vector3_type(   image_max_x , m_image_min_y,  m_image_min_z );
        m_vertices[ 2 ] = vector3_type(   image_max_x ,   image_max_y,  m_image_min_z );
        m_vertices[ 3 ] = vector3_type( m_image_min_x ,   image_max_y,  m_image_min_z );
        m_vertices[ 4 ] = vector3_type( m_image_min_x , m_image_min_y,    image_max_z );
        m_vertices[ 5 ] = vector3_type(   image_max_x , m_image_min_y,    image_max_z );
        m_vertices[ 6 ] = vector3_type(   image_max_x ,   image_max_y,    image_max_z );
        m_vertices[ 7 ] = vector3_type( m_image_min_x ,   image_max_y,    image_max_z );

        //--- Pick fixed power of two sized tiles1!!! More tiles
        //--- but easier to generate tiles and handle borders.
        unsigned int dim = math::upper_power2(  max( I, max( J, K ) ) );
        std::cout << "enclosing power 2 texture : " << dim << std::endl;

        unsigned int di = dim-2;         //--- tile dims without borders
        unsigned int dj = dim-2;
        unsigned int dk = dim-2;
        unsigned int i_width = di+2;  //--- texture dims with borders
        unsigned int j_width = dj+2;
        unsigned int k_width = dk+2;

        //--- A one-voxel large border is created around the entire volume
        //--- by shrinking it by two voxels in each axis direction.
        unsigned int i_first = 1;
        unsigned int j_first = 1;
        unsigned int k_first = 1;

        //--- compute the rest of the volume, before we hit the border
        unsigned int i_rest = (I - 1) - i_first;
        unsigned int j_rest = (J - 1) - j_first;
        unsigned int k_rest = (K - 1) - k_first;

        //--- Params needed for loading sub region of volume image into a 3D texture
        m_tile.m_image_skip_i   = (i_first-1);
        m_tile.m_image_skip_j   = (j_first-1);
        m_tile.m_image_skip_k   = (k_first-1);
        m_tile.m_image_size_i   = I;
        m_tile.m_image_size_j   = J;
        m_tile.m_image_size_k   = K;
        m_tile.m_texture_size_i = i_width;
        m_tile.m_texture_size_j = j_width;
        m_tile.m_texture_size_k = k_width;
        m_tile.m_texture_fill_i = (i_rest > i_width)?di:i_rest; //--- texutre dims are power of two, but we might not need to fill the entire texture with the volume image.
        m_tile.m_texture_fill_j = (j_rest > j_width)?dj:j_rest;
        m_tile.m_texture_fill_k = (k_rest > k_width)?dk:k_rest;

        //--- Needed for drawing AABB of tile without drawing the borders
        real_type width    = m_tile.m_texture_fill_i*dx;
        real_type height   = m_tile.m_texture_fill_j*dy;
        real_type depth    = m_tile.m_texture_fill_k*dz;
        real_type min_x    = m_image_min_x + i_first*dx;
        real_type min_y    = m_image_min_y + j_first*dy;
        real_type min_z    = m_image_min_z + k_first*dz;            
        real_type max_x    = min_x + width;
        real_type max_y    = min_y + height;
        real_type max_z    = min_z + depth;
        m_tile.m_aabb_width  = width;
        m_tile.m_aabb_height = height;
        m_tile.m_aabb_depth  = depth;
        m_tile.m_aabb_min_x  = min_x;
        m_tile.m_aabb_min_y  = min_y;
        m_tile.m_aabb_min_z  = min_z;
        m_tile.m_vertices[ 0 ] = vector3_type(   min_x ,   min_y,    min_z );
        m_tile.m_vertices[ 1 ] = vector3_type(   max_x ,   min_y,    min_z );
        m_tile.m_vertices[ 2 ] = vector3_type(   max_x ,   max_y,    min_z );
        m_tile.m_vertices[ 3 ] = vector3_type(   min_x ,   max_y,    min_z );
        m_tile.m_vertices[ 4 ] = vector3_type(   min_x ,   min_y,    max_z );
        m_tile.m_vertices[ 5 ] = vector3_type(   max_x ,   min_y,    max_z );
        m_tile.m_vertices[ 6 ] = vector3_type(   max_x ,   max_y,    max_z );
        m_tile.m_vertices[ 7 ] = vector3_type(   min_x ,   max_y,    max_z );
        m_tile.m_aabb.set( min_x, min_y, min_z, max_x, max_y, max_z);

        m_tile.m_volume_texture = OpenTissue::texture::create_texture3D<T>(
          m_tile.m_texture_size_i
          , m_tile.m_texture_size_j
          , m_tile.m_texture_size_k
          );

        m_tile.m_volume_texture->load_sub_image_into_texture(
          m_tile.m_image_skip_i
          , m_tile.m_image_skip_j
          , m_tile.m_image_skip_k
          , m_tile.m_image_size_i
          , m_tile.m_image_size_j
          , m_tile.m_image_size_k
          , 0, 0 ,0 //--- texture offset
          , m_tile.m_texture_fill_i + 2   //--- Add border
          , m_tile.m_texture_fill_j + 2
          , m_tile.m_texture_fill_k + 2
          , volume.data()
          );
        m_tile.m_volume_texture->set_repeating(false);

        //--- Needed for texture coordinate generation...
        m_tile.m_tile_width     = i_width* dx;
        m_tile.m_tile_height    = j_width* dy;
        m_tile.m_tile_depth     = k_width* dz;
        m_tile.m_tile_min_x     = m_image_min_x + (i_first-1)*dx;
        m_tile.m_tile_min_y     = m_image_min_y + (j_first-1)*dy;
        m_tile.m_tile_min_z     = m_image_min_z + (k_first-1)*dz;

        m_shader.init();
      }

      /**
      * Display the Volume.
      *
      * @param color_table_texture       The color table that should be used for displaying the volume.
      */
      template<typename texture2d_pointer>
      void display( texture2d_pointer color_table_texture)
      {
        using std::fabs;
        using std::min;
        using std::max;

        if(!color_table_texture)
          return;

        GLfloat model2eye_matrix[16];
        GLfloat eye2model_matrix[16];
        glGetFloatv( GL_MODELVIEW_MATRIX, model2eye_matrix );      
        gl::invert( model2eye_matrix, eye2model_matrix );

        m_tile.m_min_depth = math::detail::highest<real_type>();
        m_tile.m_max_depth = math::detail::lowest<real_type>();
        for(unsigned int i=0;i<8;++i)
        {
          m_tile.m_eye_vertices[i] = gl::xform( model2eye_matrix, m_tile.m_vertices[i]);
          m_tile.m_min_depth = min(m_tile.m_eye_vertices[i](2), m_tile.m_min_depth);
          m_tile.m_max_depth = max(m_tile.m_eye_vertices[i](2), m_tile.m_max_depth);      
        }
        m_shader.pre_render(  color_table_texture  );
        m_shader.render(m_tile);
        m_shader.post_render();
      }

    };

  } // namespace volume_rendering
} // namespace OpenTissue

// OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_RENDER_H
#endif
