#ifndef OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_VOLUME_SHADER_H
#define OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_VOLUME_SHADER_H
//
// OpenTissue Template Library
// - A generic toolbox for physics-based modeling and simulation.
// Copyright (C) 2008 Department of Computer Science, University of Copenhagen.
//
// OTTL is licensed under zlib: http://opensource.org/licenses/zlib-license.php
//
#include <OpenTissue/configuration.h>

namespace OpenTissue
{
  namespace volume_rendering
  {

    /**
    * The Volume Shader Class.
    * This class implements a volume shader concept by using CRTP.
    *
    * Any volume shader should be inherited from this class. As an example:
    *
    *    template<typename types>
    *    class SlabbingShader : public VolumeShader< SlabbingShader<types> >
    *
    * or
    *
    *      template<typename types>
    *      class RayCastShader : public VolumeShader< RayCastShader<types> >
    *
    * The interface is for instance used in the class TileRender.
    */
    template<typename child_type >
    class VolumeShader
    {
    public:

      /**
      * Initialization.
      * This method should be used to initialize internal data. For
      * instance setting up textures (for RTT) or frame buffer objects. It
      * could also be used for compiling and loading GPU programs.
      *
      * The method should be invoked after a render context have created.
      */
      void init()
      {
        child_type & self = static_cast<child_type>(*this);
        self.init();
      }

      /**
      * Front To Back Rendering.
      *
      * @return   If the return value is truen then tiles should be rendered
      *           in fron to back order otherwise they should be rendered in
      *           back to front order.
      */
      bool front2back() const
      {
        child_type const & self = static_cast<child_type>(*this);
        return self.front2back();
      }

      /**
      * Pre-Render.
      * This method prepares the rendering. This is a good place to perform
      * common computations that need to be done for all tiles.
      *
      * For instance retrieving the model-view transformation, clearing
      * textures (for RTT) or setting GL state. 
      */
      template<typename texture2d_pointer>
      void pre_render ( texture2d_pointer & color_table )
      {
        child_type & self = static_cast<child_type>(*this);
        self.pre_render(color_table);
      }

      /**
      * Renders a single tile.
      * Note that your shader is responsible for handling any Ping-Pong schemes inbetween tiles!
      */
      template<typename tile_type>
      void render(tile_type const & tile)    
      {
        child_type & self = static_cast<child_type>(*this);
        self.render(tile);
      }

      /**
      * Post Render.
      * This mehod is invoked after having rendered all tiles. It is usefull for
      * cleaning up GL state or transfering rendering result to frame-buffer (for RTT).
      */
      void post_render()
      {
        child_type & self = static_cast<child_type>(*this);
        self.post_render();
      }

    };

  } // namespace volume_rendering

} // namespace OpenTissue

// OPENTISSUE_UTILITY_VOLUME_RENDERING_VOLUME_RENDERING_VOLUME_SHADER_H
#endif

