// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef AUXIL_H
#define AUXIL_H

#include <string>
#include <vector>
#include <iostream>
#include "math_types.h"

/**
 Path seperator on various platforms. On Microsoft Windoes it is \, while it
 is / on other platforms. This is helpful when doing file IO in a platform
 independent manor.
 */
#ifdef _MSC_VER
const std::string path_sep ("\\");
#else
const std::string path_sep ("/");
#endif

/**
 Create a valid file name from the name of the file and the name of the directory
 containing the file. The function simply concatenates the arguments, while
 putting a platform specific path seperator between them.
 
 @param[in] dir
   The name of a directory.
 
 @param[in] filename
   The name of a file.
 
 @return
   A string containing the directory followed by a path seperator followed by the
   name of the file.
 */
inline
std::string fullfile (const std::string &dir, const std::string &filename)
{
  return dir + path_sep + filename;
}

/**
 Create a valid file name from the name of the file and the name of the directory
 containing the file. The function simply concatenates the arguments, while
 putting a platform specific path seperator between them.
 
 @param[in] dir1
   The name of a directory.
 
 @param[in] dir2
   The name of a directory.
 
 @param[in] filename
   The name of a file.
 
 @return
   A string containing the first directory followed by a path seperator followed
   the second directory followed by a path seperator and the name of the file.
 */
inline
std::string fullfile (const std::string &dir1, const std::string &dir2, const std::string &filename)
{
  return dir1 + path_sep + dir2 + path_sep + filename;
}

/**
 Find the maximal entry of a std::vector.
 
 @param[in] vec
   A std::vector of comparable elements.
   
 @return
   The index of the value with the largest value.
 */
template <class T>
inline
int find_max_idx (const std::vector<T> &vec)
{
  const int N = vec.size ();
  
  int max_idx = 0;
  
  for (int i = 1; i < N; i++)
    if (vec [i] > vec [max_idx])
      max_idx = i;
  
  return max_idx;
}

/**
 Find the maximal value of a std::vector.
 
 @param[in] vec
   A std::vector of comparable elements.
   
 @return
   The largest value.
 */
template <class T>
inline
T find_max (const std::vector<T> &vec)
{
  return vec [find_max_idx (vec)];
}

/**
 Compute the value of a number modulo two pi.
 
 @param[in] dx
   Any number.
 
 @return
   The input modulo two pi.
 */
inline
double mod_2pi (const double dx)
{
  const double dm = 2.0 * M_PI;
  if (dx < 0)
    return dm + dx + dm * floor (-dx/dm);
  else    
    return dx - dm * floor (dx/dm);
}

/**
 Compute the value of a number modulo pi.
 
 @param[in] dx
   Any number.
 
 @return
   The input modulo pi.
 */
inline
double mod_pi (const double dx)
{
  const double dm = M_PI;
  if (dx < 0)
    return dm + dx + dm * floor (-dx/dm);
  else    
    return dx - dm * floor (dx/dm);
}

/**
 Compute the square of a number.
 
 @param[in] a
   A number.
 
 @return
   The number multiplied by itself.
 */
template <class T>
inline
T square (const T &a)
{
  return a*a;
}

/**
 Compute the square two-norm of a vector.
 
 @param[in] a
   A vector.
 
 @return
   The dot product of the vector with itself (i.e. the square of its norm).
 */
template <class T>
inline
double norm2 (const T &a)
{
  return a*a;
}

/**
 Compute the two-norm of a vector.
 
 @param[in] a
   A vector.
 
 @return
   The norm of the input.
 */
template <class T>
inline
double norm (const T &a)
{
  return sqrt (norm2 (a));
}

/**
 Compute the sign of a number.
 
 @param[in] a
   Any number.
 
 @return
   If a is negative return -1, +1 or a is positive and 0 if a is 0.
 */
template <class T>
inline
T sign (const T &a)
{
  if (a < 0)
    {
      return -1;
    }
  else if (a > 0)
    {
      return 1;
    }
  else
    {
      return 0;
    }
}

/**
 Force a real number into an interval.
 
 @param[in] val
   Any real number.
 
 @param[in] min_val
   The smallest value in the interval.
 
 @param[in] max_val
   The largest value in the interval.
 
 @return
   The number nearest to val the is larger or equal to min_val and smaller or
   equal to max_val.
 */
template<class T>
inline
T force_in_range (const T &val, const T &min_val, const T &max_val)
{
  if (val > max_val)
    {
      return max_val;
    }
  else if (val < min_val)
    {
      return min_val;
    }
  else
    {
      return val;
    }
} 

/**
 Project a 3D point onto a 3D line.
 
 @param[in] p
   The 3D point to be projected onto a line.
 
 @param[in] l1
   One point on the line.
 
 @param[in] l2
   Another point on the line (different from l1)
 
 @return
   The projection of p into the line spanned by l1 and l2.
 */
inline
vector3 project_point_on_line (const vector3 &p, const vector3 &l1, const vector3 &l2)
{
  // XXX: This is a silly implementation
  const vector3 delta = l2 - l1;
  const double t = - ((l1 - p) * delta) / norm2 (delta);
  const vector3 retval = l1 + delta * t;
  
  return retval;
}

/**
 The Jet colour map ranging form blue to red.
 
 @param[in] x
   A real number between 0 and 1.
 
 @return
   A 3-vector containing RGB values corresponding to the input.
 */
vector3 jet (const double x);

/**
 The Hot colour map ranging form black to white.
 
 @param[in] x
   A real number between 0 and 1.
 
 @return
   A 3-vector containing RGB values corresponding to the input.
 */
vector3 hot (const double x);

/**
 Compute the Modified Bessel function of order 0.
 
 @param[in] x
   A real number.
 
 @return
   The function value of the modified Bessel function of order 0.
 */
double bessi0 (double x);

#endif // AUXIL_H
