// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef BROWNIAN_STATE_H
#define BROWNIAN_STATE_H

#include "angle_state.h"
#include "options.h"
#include "simulator.h"
#include "auxil.h"

#include <OpenTissue/core/math/big/big_svd.h>
#include <boost/numeric/ublas/banded.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <iostream>

/**
 An implementation of the spatial Brownian motion prior. If you use this class we recommend
 citing
 @code
 @article{hauberg:imavis11,
  title = {{Natural Metrics and Least-Committed Priors for Articulated Tracking}},
  author = {S{\o}ren Hauberg and Stefan Sommer and Kim Steenstrup Pedersen},
  journal = {Image and Vision Computing},
  publisher = {Elsevier},
  year = {2011},
 }
 @endcode
 */
template <class observation_type>
class brownian_state : public angle_state<observation_type>
{
public:
  typedef boost::numeric::ublas::diagonal_matrix<double> diagonal_matrix_type;

  brownian_state (const options &opts, const calibration &_calib, const bool _for_show = true)
    : angle_state<observation_type> (opts, _calib, _for_show),
      pred_noise (opts.pred_noise ())
    {
      this->compute_pose ();
  
      // Store default chains temporarely
      std::list<vector3> chain_list;
      for (chain_iter iter = this->solver.chain_begin (); iter != this->solver.chain_end (); iter++)
        chain_list.push_back (iter->get_end_effector ()->absolute ().T ());
  
      // Add everything as end-effectors
      int i = 0;
      for (bone_iter iter = this->bone_begin (); iter != this->bone_end (); iter++, i++)
        {
          if (iter->is_root ())
            continue;
        
          const vector3 absolute = iter->absolute ().T ();
          const vector3 pabsolute = iter->parent ()->absolute ().T ();

          // Make sure the bone isn't already an end-effector
          bool accept = true; //(absolute != pabsolute);
          for (std::list<vector3>::const_iterator ci = chain_list.begin ();
               ci != chain_list.end (); ci ++)
            {
              if (*ci == absolute)
                {
                  accept = false;
                  break;
                }
            }
          
          if (accept)
            {
              chain_type chain;
              chain.init (this->skeleton.root (), &(*iter));
              this->solver.add_chain (chain);
            }
        }
      
      // Count number of end-effectors
      dim = 0;
      for (chain_iter iter = this->solver.chain_begin (); iter != this->solver.chain_end (); iter++)
        dim += 3;

      initial_pose.resize (dim);
      int k = 0;
      for (chain_iter iter = this->solver.chain_begin (); iter != this->solver.chain_end (); iter++)
        {
          const vector3 e = iter->get_end_effector ()->absolute ().T ();
          for (size_t n = 0; n < 3; n++)
            initial_pose (k++) = e (n);
        }
    }

private:
  void jacobian (matrix_type &J)
    {
      for (chain_iter iter = this->solver.chain_begin (); iter != this->solver.chain_end (); iter++)
        iter->p_global () = iter->get_end_effector ()->absolute ().T ();

      compute_jacobian (this->solver.chain_begin (), this->solver.chain_end (),
                        this->bone_begin (), this->bone_end (), J);
    }
  
  void current_pos (vector_type &x)
    {
      int idx = 0;
      for (chain_iter iter = this->solver.chain_begin (); iter != this->solver.chain_end (); iter++)
        {
          vector3 local_x = iter->get_end_effector ()->absolute ().T ();
          for (int d = 0; d < 3; d++)
            x (idx++) = local_x (d) + this->root (d); // XXX: should we add root?
        }
    }

  void project (vector_type &x)
    {
      int idx = 0;
      for (chain_iter iter = this->solver.chain_begin (); iter != this->solver.chain_end (); iter++)
        {
          vector3 local_x;
          for (int d = 0; d < 3; d++)
            local_x (d) = x (idx++);
                
          iter->p_global () = local_x - this->root;
          
          // XXX: Check to ensure no ground penetrations
        }

      solver_type::Output output;
      this->solver.solve (&solver_type::default_SD_settings (), &output);
      OpenTissue::kinematics::inverse::get_joint_parameters (*(this->solver.skeleton ()),
                                                             this->theta);
      //std::cout << output.value () << " " << output.gradient_norm () << std::endl;
    }

public:
  // Implementation of the Euler-Heun scheme (works with Stratonovich models)
  double predict (const observation_type &observation, const double variance_scale = 1.0)
    {
      const double std [] = {0.019905, 0.003157, 0.008050, 0.180706, 0.101031, 0.137931,
                             0.275119, 0.182599, 0.286416, 0.275119, 0.182599, 0.286416,
                             0.390496, 0.301062, 0.483356, 0.508283, 0.808703, 0.806479,
                             0.275119, 0.182599, 0.286416, 0.381652, 0.255074, 0.218130,
                             0.498532, 0.249264, 0.615678, 0.938598, 0.593349, 1.098928,
                             0.275119, 0.182599, 0.286416, 0.509799, 0.464794, 0.370578,
                             0.988121, 0.796506, 0.450776, 1.155469, 1.320700, 0.709060};
    
      const int num_time_steps = 10;
      const double normalisation = sqrt ((double)num_time_steps);
    
      matrix_type J; // 30x41
      ublas::matrix<double> U, V, UUT;
      vector_type S, W1 (dim), W2 (dim), incr1 (dim), incr2 (dim), x (dim), x_tmp (dim);
      diagonal_matrix_type Sd;
      
      gaussian1D g;
      
      // Predict root here if needed

      for (int time_step = 0; time_step < num_time_steps; time_step++)
        {
          // Simulate Euclidean Brownian motion
          for (size_t k = 0; k < dim; )
            {
              for (size_t n = 0; n < 3; n++,  k++)
                {
                  const double s = std [k] / normalisation;
                  W1 (k) = s * g.random ();
                  W2 (k) = s * g.random ();
                }
              const double s = ((k >= 21) ? 3.0 * pred_noise :  pred_noise) / normalisation; 
              for (size_t n = 0; n < 3; n++,  k++)
                {
                  W1 (k) = s * g.random (); // - this->root (n);
                  W2 (k) = s * g.random (); // - this->root (n);
                }
            }

          // Get current position          
          this->compute_pose ();
          current_pos (x);

          // Compute P = U U^T at current position
          jacobian (J);
          OpenTissue::math::big::svd (J, U, S, V);
          if (Sd.size1 () == 0)
            Sd.resize (S.size (), S.size ());
          Sd.clear ();
          for (size_t i = 0; i < S.size (); i++)
            Sd (i, i) = (S (i) > 1e-9) ? 1.0 : 0.0;
          U = prod (U, Sd);
          UUT = prod (U, trans (U));

          // Compute the ordinary Euler step
          incr1 = prod (UUT, W1);

          // Perform inner Euler step
          incr2 = prod (UUT, W2);
          x_tmp = x + incr2;
          project (x_tmp);
          jacobian (J);
          OpenTissue::math::big::svd (J, U, S, V);

          //Sd.resize (S.size (), S.size ());
          Sd.clear ();
          for (size_t i = 0; i < S.size (); i++)
            Sd (i, i) = (S (i) > 1e-9) ? 1.0 : 0.0;

          U = prod (U, Sd);
          UUT = prod (U, trans (U));
          incr2 = prod (UUT, W1);
          
          // Euler-Heun Step
          x = x + 0.5 * (incr1 + incr2);
          project (x);
        }
        
      return 0;
    }

  brownian_state& operator= (brownian_state &new_state)
    {
      angle_state<observation_type>::operator= (new_state);
      dim = new_state.dim;

      return *this;
    }
    
  bool load (const std::string filename)
    {
      return angle_state<observation_type>::load (filename);
    }

  bool load (std::ifstream &file)
    {
      return angle_state<observation_type>::load (file);
    }

  private:
    const double pred_noise;
    size_t dim;
    vector_type initial_pose;
};

#endif // BROWNIAN_STATE_H
