// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef CALIBRATION_H
#define CALIBRATION_H

#include "options.h"
#include "cxcore.h"
#include <string>

class calibration
{
public:
  /**
   Constructor for the calibration class. This class allows you to change
   from the different coordinate systems. In general we work with three coordinate
   systems:
   <ul>
     <li>The world coordinate system. This 3D coordinate system is the one in which
       the actual tracking is performed. Most data is also represented internally
       in this system. The actual definition of this coordinate system may vary,
       but often the origin is placed in the root of the kinematic skeleton.</li>
     <li>The camera coordinate system. This 3D coordinate system has the optical
       centre as the origin. The Z-axis is perpendicular to the image plane.</li>
     <li>The image coordinate system. This 2D coordinate system corresponds to
       the pixel coordinates of the image.</li>
   </ul>
   
   @param[in] opts
     An object of class options that contains the information in a configuration file.
     The constructor reads the "intrinsic_matrix" and "external_view" options from
     opts.
   */
  calibration (const options &opts)
    : intrinsic (opts.intrinsic_matrix ())
    {
      load_world (opts.external_view ());
    }
  
  const matrix3& get_intrinsic_matrix () const
    {
      return intrinsic;
    }
  
  /**
   Project a 3D point in the camera coordinate system onto the image. The image
   point is optionally scaled.
   
   @param[in] v
     The 3D point expressed in the camera coordinate system.
   @param[in] scale
     An optional scaling factor (defaults to 1). The resulting pixel coordinate
     is multiplied with this factor. This is useful when projecting points into
     an image of different size than the original image.
   
   @return
     The pixel coordinate of the projected point.
   */
  CvPoint camera_to_pixel (const vector3 &v, double scale = 1.0) const
    {
      CvPoint retval;
      
      const vector3 p = intrinsic * v;
      // We add 0.5 to ensure pixel coordinates are rounded rather than floored (retval contains integers).
      retval.x = scale * p (0) / p (2) + 0.5;
      retval.y = scale * p (1) / p (2) + 0.5;
      
      return retval;
    }
  
  /**
   Project a 3D point in the world coordinate system onto the image. The image
   point is optionally scaled.
   
   @param[in] v
     The 3D point expressed in the world coordinate system.
   @param[in] scale
     An optional scaling factor (defaults to 1). The resulting pixel coordinate
     is multiplied with this factor. This is useful when projecting points into
     an image of different size than the original image.
   
   @return
     The pixel coordinate of the projected point.
   */
   CvPoint world_to_pixel (const vector3 &v, double scale = 1.0) const
    {
      const vector3 c = world_to_camera (v);
      return camera_to_pixel (c, scale);
    }
  
  /**
   Convert a 2D pixel coordinate to a 3D direction expressed in the camera coordinate
   system.
   
   @return
     A 3D vector pointing from the origin through the pixel. The Z-coordinate
     of this vector is 1.
   */
  vector3 pixel_to_camera (const CvPoint p, double scale = 1.0) const
    {
      vector3 retval;
      retval [0] = (((double)p.x) / scale - intrinsic (0, 2)) / intrinsic (0, 0);
      retval [1] = (((double)p.y) / scale - intrinsic (1, 2)) / intrinsic (1, 1);
      retval [2] = 1.0;
      return retval;
    }
  
  /**
   Convert a 2D pixel coordinate to a 3D direction expressed in the world coordinate
   system.
   
   @return
     A 3D vector pointing from the origin through the pixel.
   */
  vector3 pixel_to_world (const CvPoint p, double scale = 1.0) const
    {
      const vector3 c = pixel_to_camera (p, scale);
      return camera_to_world (c);
    }
  
  /**
   Change a point from the world coordinate system to the camera coordinate system.
   
   @param[in] v
     A 3D point expressed in the world coordinate system.
   
   @return
     A 3D point expressed in the camera coordinate system.
   */
  vector3 world_to_camera (const vector3 &v) const
    {
      // XXX: 100 is a hack!!
      const vector3 retval = 100.0 * (inv_world_R * v - world_translation);
      return retval;
    }
  
  /**
   Change a point from the camera coordinate system to the world coordinate system.
   
   @param[in] v
     A 3D point expressed in the camera coordinate system.
   
   @return
     A 3D point expressed in the world coordinate system.
   */
  vector3 camera_to_world (const vector3 &v) const
    {
      // XXX: 100 is a hack!!
      const vector3 retval = world_R * (v / 100.0 + world_translation);
      return retval;
    }
  
  /**
   Return the origin of the camera coordinate system represented in the world
   coordinate system. This simply corresponds to the following code.
   @code
   camera_to_world (vector3 (0, 0, 0));
   @endcode
   
   @return
     A 3D point corresponding to the origin of the camera represented in the
     world coordinate system.
   */
 vector3 camera_centre () const
   {
     return camera_to_world (vector3 (0, 0, 0));
   }
    
  /**
   Set the translation parameters for the transformation from the
   camera coordinate system to the world coordinate system.
     
   @param[in] t
     The translation vector from the camera coordinate system to the world
     coordinate system.
   */
  void set_world_translation (const vector3 &t)
    {
      world_translation = t;
    };
      
  /**
   Return the current translation parameters for the transformation between the
   camera coordinate system and the world coordinate system.
   
   @return
     The translation vector from the camera coordinate system to the world
     coordinate system.
   */
  vector3 get_world_translation () const
    {
      return world_translation;
    };
      
  /**
   Set the rotation parameters for the transformation from the
   camera coordinate system to the world coordinate system.
     
   @param[in] R
     A 3-vector where each element correspond to a rotation angle around one
     coordinate axis.
   */
  void set_world_rotation (const vector3 &R)
    {
      world_rot_vec = R;
  
      const matrix3 rxm = OpenTissue::math::Rx (R [0]);
      const matrix3 rym = OpenTissue::math::Ry (R [1]);
      const matrix3 rzm = OpenTissue::math::Rz (R [2]);
  
      world_R = rxm * rym * rzm;

      const matrix3 irxm = OpenTissue::math::Rx (-R [0]);
      const matrix3 irym = OpenTissue::math::Ry (-R [1]);
      const matrix3 irzm = OpenTissue::math::Rz (-R [2]);
  
      inv_world_R = irzm * irym * irxm;
    };

  /**
   Return the current rotation parameters for the transformation from the
   camera coordinate system to the world coordinate system.
   
   @return
     A 3-vector where each element correspond to a rotation angle around one
     coordinate axis.
   */
  vector3 get_world_rotation () const
    {
      return world_rot_vec;
    };

  /**
   Save the current translation and rotation paramters for the transformation
   from the camera coordinate system to the world coordinate system.
     
   @param[in] filename
     The name of the file in which the parameters will be stored.
     
   @return
     The function returns true of the data was written succesfully and false otherwise.
   */
  bool save_world (const std::string filename) const
    {
      bool retval = false;
      std::ofstream rtfile (filename.c_str ());
      if (rtfile.is_open ())
        {
          rtfile << world_translation [0] << " " << world_translation [1] << " "
                 << world_translation [2] << std::endl;
          rtfile << world_rot_vec [0] << " " << world_rot_vec [1] << " "
                 << world_rot_vec [2] << std::endl;
          rtfile.close ();
          retval = true;
        }
      else
        std::cerr << "Could not open '" << filename << "' for writing" << std::endl;
    
      return retval;
    };

  /**
   Read translation and rotation parameters for the transformation from the
   camera coordinate system to the world coordinate system.
     
   @param[in] filename
     The name of the file containing the translation and rotation parameters.
     The file should contain one single line of ASCII text containing 6 numbers.
     These numbers should be
     <ul>
       <li>The translation along the X-axis.</li>
       <li>The translation along the Y-axis.</li>
       <li>The translation along the Z-axis.</li>
       <li>The rotation around the X-axis.</li>
       <li>The rotation around the Y-axis.</li>
       <li>The rotation around the Z-axis.</li>
     </ul>
     
   @return
     The function return true if the file was read succesfully and false otherwise.
   */
  bool load_world (const std::string filename)
    {
      bool retval = false;
      std::ifstream rtfile (filename.c_str ());
      if (rtfile.is_open ())
        {
          for (size_t k = 0; k < 3; k++)
            rtfile >> world_translation [k];
          for (size_t k = 0; k < 3; k++)
            rtfile >> world_rot_vec [k];
          rtfile.close ();
          set_world_rotation (world_rot_vec);
          retval = true;
        }
      else
        std::cerr << "Could not open '" << filename << "' for reading" << std::endl;
    
      return retval;
    };

private:
  matrix3 intrinsic, world_R, inv_world_R;
  vector3 world_translation, world_rot_vec;
};

#endif // CALIBRATION_H

