// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef CLOUD_H
#define CLOUD_H

#include <string>
#include <vector>
#include <iostream>
#include <fstream>

#include "observation.h"
#include "options.h"
#include "math_types.h"
#include "calibration.h"
#include "project2d.h"
#include "opencv_helper.h"
#include "knn_cloud.h"

/**
 An abstract class for representing data from a stereo camera. The template parameter
 controls the type of observation the camera generates. Examples of such observations
 include 3D points and triangles in 3D.
 
 The data should internally be represented in the world coordinate system rather
 than the camera coordinate system.
 */
template <class observation_type>
class cloud : public observation<observation_type>
{
  public:
    /**
     Constructor for the cloud class. The constructor initialises some
     variables, but does not actually read any data.
     
     @param[in] opts
       An object of class options that contains configuration information. The
       cloud constructor reads the following options:
       <ul>
         <li>data_template</li>
         <li>data_start_idx</li>
         <li>external_view</li>
         <li>data_scaling (optional)</li>
         <li>depth_threshold (optional)</li>
       </ul>
     
     @param[in] _calib
       A reference to the calibration object defining the coordinate systems of
       the world.
     */
    cloud (const options &opts, const calibration &_calib)
      : frame (opts.data_start_idx ()),
        scale (opts.data_scaling ()),
        calib (_calib)
      {
        try
          {
            depth_threshold = opts.depth_threshold ();
            use_depth_threshold = true;
          }
        catch (missing_option)
          {
            use_depth_threshold = false;
          }
      };
    
    ~cloud () {}; // Just here to remove a warning
    
    /**
     Reads data from the next frame. This function should be implemented by a class
     inheriting from the cloud class.

     @param[in] goto_next
       This optional parameter determines if the next frame should be read (default)
       or the current frame should be re-read. The latter is useful if you have
       changed parameters of the point_cloud object, such as the translation or
       rotation of the data.
     
     @return
       If the frame was succesfully read the function returns true, otherwise it
       returns false.
     */
    virtual bool refresh (bool goto_next = true) = 0;
    
    /**
     @return
       Returns the number of observation elements in the currently held frame.
     */
    virtual size_t size () const = 0;

    /**
     Draw the data currently held in the object. This function must be implemented
     by a class inheriting from cloud.
     */
    virtual void show () const = 0;
  
  protected:
    int frame;
    const double scale;
    bool use_depth_threshold;
    double depth_threshold;
    const calibration &calib;
};

/**
 The point_vector class corresponds to an array of 3D points. This is mostly used
 to represent 3D points generated by a stereo camera.
 */
typedef std::vector<vector3> point_vector;

/**
 cloud3 is an abstract class used for representing observations consisting of
 coloured 3D points, such as those generated by stereo cameras or depth sensors.
 */
class cloud3 : public cloud<point_vector>
{
  public:
    /**
     Constructor for the cloud3 class. Call this from derived class constructors.
     Also, do not forgot to call the 'init' function as part of the derived
     constructor.
     */
    cloud3 (const options &opts, const calibration &_calib)
      : cloud<point_vector> (opts, _calib),
        downsample (opts.measure_downsampling ()),
        first_frame (frame)
      {
        background = cvCreateImage (cvSize (PROJ2D_SCALE * 1024.0, PROJ2D_SCALE * 768.0), IPL_DEPTH_8U, 1);
        try
          {
            depth_background = opts.depth_background ();
            use_background_subtraction = true;
          }
        catch (missing_option)
          {
            use_background_subtraction = false;
          }
      }
    
    /**
     Initialise the cloud3 object. You need to call this in the constructor of
     derived classes.
     */
    void init ()
      {
        if (use_background_subtraction)
          use_background_subtraction = read_background (depth_background);
      }
    
    /**
     Attempt to read information about the scene background. You need to
     implement this function in derived classes.
     
     @return
       true if the background could be read succesfully; false otherwise.
     */
    virtual bool read_background (const std::string &depth_background)
      {
        std::cerr << "warning: 'read:background' not implemented in 'cloud3'" << std::endl;
        return false;
      }
  
    /**
      Render background subtracted image from the 3D point. Each 3D point currently
      stored in the object is projected into retval (using the calibration object
      given to the object constructor). Projected pixels are rendered
      with a value of 255, while other pixels are assigned a 0 value. A morphologic
      closing operation is performed on the image afterwards.
     
      @param[out] retval
        An 8-bit unsigned gray scale image (IPL_DEPTH_8U with 1 channel) in which
        the result is rendered.
      
      @param[in] scale
        An optional scaling parameter. This is passed to the world_to_pixel function
        in the calibration object.
     */
   void render_background_image (IplImage *retval, double scale = 1.0) const
     {
        cvZero (retval);
        const size_t num_points = points.size ();
        for (size_t k = 0; k < num_points; k++)
          {
            const CvPoint pixel = calib.world_to_pixel (points [k], scale);
            GRAY (retval, pixel.x, pixel.y) = 255;
          }
        cvMorphologyEx (retval, retval, NULL, NULL, CV_MOP_CLOSE);
     }
   
    /**
     Draw the point cloud using OpenGL.
     */
    void show () const
      {
        glPointSize (4.5f);
        //glPointSize (6.0f);

        glDisable (GL_LIGHTING);
        glBegin (GL_POINTS);
        for (size_t k = 0; k < points.size (); k++)
          {
            const vector3 rgb = colours [k];
            const vector3 p = points [k];
            glColor3f (rgb (0), rgb (1), rgb (2));
            glVertex3f (p [0], p [1], p [2]);
          }
        glEnd ();
        glEnable (GL_LIGHTING);
      }

    /**
     Return one 3D point. This allows you to access the data currently held in the
     object. As an example,
     @code
     for (size_t i = 0; i < my_point_cloud.size (); i++)
       do_something_with_point (my_point_cloud [i]);
     @endcode
     
     @param[in] i
       The index of the point you want to read.
       
     @return
       The 3D point at the given index.
     */
    vector3 operator[] (const size_t i) const
      {
        return points [i];
      };

    /**
     @return
       The number of 3D points in the current frame.
     */
    size_t size () const
      {
        return points.size ();
      }

   /**
    Return a project2d object where the binary segment corresponds to a projection
    of the 3D points onto the image plane. This allows for easy 2D projection of
    2D points onto the binary segment.
    
    @return
      An object of class project2d. The used binary segment is a projection of
      the 3D points onto the original image plane.
    */
   const project2d& get_project2d () const
     {
       return proj2d;
     }
       
    /**
     Return a knn_cloud object corresponding to the 3D points in the point_cloud.
     
     @return
       An object of class knn_cloud.
     */
   const knn_cloud& get_knn_cloud () const
     {
       return nearest_cloud;
     }

    /**
     Return a knn_cloud object corresponding to the 3D points in the background.
     
     @return
       An object of class knn_cloud.
     */
   const knn_cloud& get_knn_bg_cloud () const
     {
       return bg_nearest_cloud;
     }

  protected:
    point_vector points, colours;
    const int downsample;
    const int first_frame;
    bool use_background_subtraction;
    std::string depth_background;
    project2d proj2d;
    knn_cloud nearest_cloud, bg_nearest_cloud;
    IplImage *background;
};
#endif // CLOUD_H
