// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.

#ifdef HAVE_TRICLOPS

#include "cvBumblebee.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <assert.h>

#include "index.h"
#include "auxil.h"

#include <iostream>

// TODO: We should use exceptions instead of just calling exit

// Constructor / Destructor
cvBumblebee::cvBumblebee (const std::string &conf)
  : ipl_left (NULL),
    ipl_right (NULL),
    ipl_disparity (NULL)
{
  allocate_buffers (cvSize (1024, 768));

  char config_file [conf.size ()+1];
  strcpy (config_file, conf.c_str ());

  // Prepare for stereo processing
  e = triclopsGetDefaultContextFromFile (&triclops, config_file);
  if (e != TriclopsErrorOk)
    {
      std::cerr << "Can't get context from configuration file" << std::endl;
      exit (1);
    }
  
  set_stereo_parameters ();
  
  float focal_length, row, col;
  triclopsGetFocalLength (triclops, &focal_length);
  triclopsGetImageCenter (triclops, &row, &col);
}

void cvBumblebee::set_stereo_parameters ()
{
  // Make sure we are in subpixel mode (XXX: which options should we enable)
  //trisize = cvSize (1024, 768);
  trisize = cvSize (640, 480);
  //trisize = cvSize (320, 240);
  triclopsSetSubpixelInterpolation (triclops, 1);
  triclopsSetResolution (triclops, trisize.height, trisize.width);
  triclopsSetDisparity (triclops, 10, 100);
  triclopsSetTextureValidation (triclops, 1);
  triclopsSetUniquenessValidation (triclops, 1);
  triclopsSetEdgeMask (triclops, 11);
  std::cerr << __LINE__ << std::endl;
  triclopsSetMaxThreadCount (triclops, 1); // only use one thread for stereo to avoid crash in Triclops
  std::cerr << __LINE__ << std::endl;
}

void cvBumblebee::allocate_buffers (const CvSize size)
{
  // size of buffer for all images at mono8
  const unsigned int num_pix = size.width * size.height;
   
  offline_input.nrows = size.height;
  offline_input.ncols = size.width;
  offline_input.rowinc = size.width;
  offline_input.inputType = TriInp_RGB;
  offline_input.u.rgb.red = (void*) (new unsigned char [num_pix]);
  offline_input.u.rgb.green = (void*) (new unsigned char [num_pix]);
  offline_input.u.rgb.blue = (void*) (new unsigned char [num_pix]);

  ipl_left = cvCreateImage (size, IPL_DEPTH_8U, 3);
  ipl_right = cvCreateImage (size, IPL_DEPTH_8U, 3);
}

cvBumblebee::~cvBumblebee ()
{
  if (ipl_left)
    cvReleaseImage (&ipl_left);
  if (ipl_right)
    cvReleaseImage (&ipl_right);
}

CvSize cvBumblebee::get_depth_size () const
{
  return trisize;
}

IplImage* cvBumblebee::get_disparity ()
{
  triclopsSetResolution (triclops, trisize.height, trisize.width);
  e = triclopsStereo (triclops);
  if (e != TriclopsErrorOk)
    {
      std::cerr << "triclopsStereo failed!" << std::endl;
      triclopsDestroyContext (triclops);
      return NULL;
    }

  // get the rectified and disparity images
  triclopsGetImage16 (triclops, TriImg16_DISPARITY, TriCam_REFERENCE, &image16);

  ipl_disparity = cvCreateImage (trisize, IPL_DEPTH_16U, 1);

  const int pixelinc = image16.rowinc / 2;
  for (int i = 0; i < image16.nrows; i++)
    {
      unsigned short *row = image16.data + i * pixelinc;
      for (int j = 0; j < image16.ncols; j++)
        {
          const unsigned short disparity = row [j];
          GRAY16 (ipl_disparity, j, i) = disparity;
        }
    }
  
  return ipl_disparity;
}

vector3 cvBumblebee::get_xyz (int x, int y, bool &valid)
{
  const int pixelinc = image16.rowinc / 2;
  unsigned short *row = image16.data + y * pixelinc;
  const unsigned short disparity = row [x];
  return get_xyz (disparity, x, y, valid);
}

vector3 cvBumblebee::get_xyz (unsigned short disparity, int x, int y, bool &valid)
{
  float vx, vy, vz;
  
  if (invalid_disparity (disparity))
    valid = false;
  else
    {
      triclopsRCD16ToXYZ (triclops, y, x, disparity, &vx, &vy, &vz);
      valid = true;
    }
    
  const double s = 1000.0;
  return vector3 (s*vx, -s*vy, s*vz);
}

void cvBumblebee::set_images (IplImage *left, IplImage *right)
{
  // 1. Build TriclopsInputRGB with channels (left, right, left)
  ipl_to_buffer (right, (unsigned char*)offline_input.u.rgb.red);
  ipl_to_buffer (left, (unsigned char*)offline_input.u.rgb.green);
  ipl_to_buffer (left, (unsigned char*)offline_input.u.rgb.blue); // no 3rd camera

  // 2. Call preprocess function with input to ensure that context is updated
  e = triclopsPreprocess (triclops, &offline_input);
  if (e != TriclopsErrorOk)
    std::cerr << "triclopsPreprocess failed!" << std::endl;
}

void cvBumblebee::ipl_to_buffer (IplImage *ipl, unsigned char *buffer)
{
  const CvSize size = cvGetSize (ipl);
  if (ipl->nChannels == 3)
    {
      for (int y = 0; y < size.height; y++)
        {
          for (int x = 0; x < size.width; x++)
            {
              const int val = ((int)RED (ipl, x, y) + (int)GREEN (ipl, x, y) +
                               (int)BLUE (ipl, x, y)) / 3;
              *buffer = val;
              buffer ++;
            }
        }
    }
  else
    {
      for (int y = 0; y < size.height; y++)
        {
          for (int x = 0; x < size.width; x++)
            {
              *buffer = GRAY (ipl, x, y);
              buffer++;
            }
        }
    }
}

#endif // HAVE_TRICLOPS

