// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.

#ifdef HAVE_TRICLOPS
#ifndef CV_BUMBLEBEE_H
#define CV_BUMBLEBEE_H

#include "triclops.h"
#include "cxcore.h"
#include "cv.h"
#include "highgui.h"
#include "math_types.h"

#define NO_DISPARITY 0xFF00

inline bool invalid_disparity (unsigned short disparity)
{
  return (disparity >= NO_DISPARITY);
}

// TODO: We should use exceptions instead if just returning

/**
 A simple class for getting 3D points from a Bumblebee camera. For this we use
 the Triclops library that comes with the camera.
 */
class cvBumblebee
{
public:
  /**
   @param [in] conf
   The configuration file for the Triclops library.
   */
  cvBumblebee (const std::string &conf);
  ~cvBumblebee ();
  
  /**
   Compute disparity.
   
   @return
   A 16 bit gray-scale image containing the disparity.
   */
  IplImage* get_disparity ();
  
  /**
   Extract (x, y, z) coordinates of a given pixel.
   
   @param [in] x
   The pixel x-corodinate.
   
   @param [in] y
   The pixel y-coordinate.
   
   @param [out] valid
   Was it possible to reconstruct the pixel in 3D.
   
   @return
   The 3D coordinates of the pixel.
   */
  vector3 get_xyz (int x, int y, bool &valid);

  /**
   Extract (x, y, z) coordinates of a given pixel.
   
   @param [in] disparity
   The disparity of the pixel.
   
   @param [in] x
   The pixel x-corodinate.
   
   @param [in] y
   The pixel y-coordinate.
   
   @param [out] valid
   Was it possible to reconstruct the pixel in 3D.
   
   @return
   The 3D coordinates of the pixel.
   */
  vector3 get_xyz (unsigned short disparity, int x, int y, bool &valid);
  
  /**
   Set the left and right images for the stereo reconstruction.
   
   @param [in] left
   The left image.
   
   @param [in] right
   The right image.
   */
  void set_images (IplImage *left, IplImage *right);
  
  /**
   @return
   The size of the disparity image.
   */
  CvSize get_depth_size () const;

private:
  // Data
  IplImage *ipl_left, *ipl_right, *ipl_disparity;
  TriclopsImage16 image16;
  
  TriclopsInput offline_input;
  TriclopsError e;
  TriclopsContext triclops;
  CvSize trisize;
  
protected:
  void allocate_buffers (const CvSize size);
  void set_stereo_parameters ();
  void ipl_to_buffer (IplImage *ipl, unsigned char *buffer);
};

#endif // CV_BUMBLEBEE_H
#endif // HAVE_TRICLOPS
