// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef ELLIPSOID_H
#define ELLIPSOID_H

#include "math_types.h"
#include <OpenTissue/utility/gl/gl_draw_ellipsoid.h>

/**
 Different implementations of how to project a point onto a ellipsoid.
 */
enum ELLIPSOID_PROJECTION_TYPE
{
  MAHALANOBIS
};

/**
 A basic class for representing ellipsoids in 3D.
 */
class ellipsoid
{
public:
  typedef double real_type;
  typedef vector3 vector3_type;

  /**
   Default constructor that assigns radius 0 to the ellipsoid
   */
  ellipsoid ()
    : r0 (0), r1 (0), r2 (0)
    {
    };

  /**
   @param[in] C
   The 3D point corresponding to the center of the ellipsoid.
 
   @param[in] a0
   The 3D vector corresponding to the first principal axis of the ellipsoid.
 
   @param[in] a1
   The 3D vector corresponding to the second principal axis of the ellipsoid.
 
   @param[in] a2
   The 3D vector corresponding to the third principal axis of the ellipsoid.
 
   @param[in] _r0
   The length of the first principal axis.
 
   @param[in] _r1
   The length of the first principal axis.
 
   @param[in] _r2
   The length of the first principal axis.
   */
  ellipsoid (vector3 C, vector3 a0, vector3 a1, vector3 a2, double _r0, double _r1, double _r2)
    : m_center (C),
      m_axis0 (a0),
      m_axis1 (a1),
      m_axis2 (a2),
      r0 (_r0),
      r1 (_r1),
      r2 (_r2)
    {
    };
  
  /**
   Initialises the parameters of an ellipsoid. Useful when the ellipsoid has been
   created using the default constructor.
   
   @param[in] C
   The 3D point corresponding to the center of the ellipsoid.
 
   @param[in] a0
   The 3D vector corresponding to the first principal axis of the ellipsoid.
 
   @param[in] a1
   The 3D vector corresponding to the second principal axis of the ellipsoid.
 
   @param[in] a2
   The 3D vector corresponding to the third principal axis of the ellipsoid.
 
   @param[in] _r0
   The length of the first principal axis.
 
   @param[in] _r1
   The length of the first principal axis.
 
   @param[in] _r2
   The length of the first principal axis.
   */
  void init (vector3 C, vector3 a0, vector3 a1, vector3 a2, double _r0, double _r1, double _r2)
    {
      m_center = C;
      m_axis0 = a0;
      m_axis1 = a1;
      m_axis2 = a2;
      r0 = _r0;
      r1 = _r1;
      r2 = _r2;
    }
  
  /**
   Project a point onto the ellipsoid.
   
   @param [in] input
   The point to be projected.
   
   @param [out] output
   The result of the projection.
   
   @param [in] proj_type
   The choice of projection type.
   */
  void project_point (const vector3 &input, vector3 &output, ELLIPSOID_PROJECTION_TYPE proj_type = MAHALANOBIS) const
    {
      double N;
      project_point (input, output, N, proj_type);
    }
    
  /**
   Project a point onto the ellipsoid.
   
   @param [in] input
   The point to be projected.
   
   @param [out] output
   The result of the projection.
   
   @param [out] N
   
   @param [in] proj_type
   The choice of projection type.
   */
  void project_point (const vector3 &input, vector3 &output, double &N, ELLIPSOID_PROJECTION_TYPE proj_type = MAHALANOBIS) const
    {
      if (proj_type == MAHALANOBIS)
        {
          const vector3 delta = input - m_center;
          N = square (dot (delta, m_axis0))/r0 + square (dot (delta, m_axis1))/r1 + square (dot (delta, m_axis2))/r2;
          output = delta/N + m_center;
        }
    }
  
  double euc_mahalanobis2 (const vector3 &p) const
    {
      return euc_mahalanobis2 (p, m_center);
    }

  double euc_mahalanobis2 (const vector3 &p, const vector3 &center) const
    {
      const vector3 delta = p - center;
      const double dist2 = square (dot (delta, m_axis0)) / r0
                         + square (dot (delta, m_axis1)) / r1
                         + square (dot (delta, m_axis2)) / r2;
      return dist2;
    }
  
  void show () const
    {
      OpenTissue::gl::ColorPicker (0.7, 0.7, 0.7, 0.5);
      OpenTissue::gl::DrawEllipsoid<ellipsoid> (*this);
    }
    
  const vector3& center () const
    {
      return m_center;
    };
    
  const vector3& axis0 () const
    {
      return m_axis0;
    };

  const vector3& axis1 () const
    {
      return m_axis1;
    };

  const vector3& axis2 () const
    {
      return m_axis2;
    };

  const double& radius0 () const
    {
      return r0;
    };

  const double& radius1 () const
    {
      return r1;
    };

  const double& radius2 () const
    {
      return r2;
    };

private:
  vector3 m_center, m_axis0, m_axis1, m_axis2;
  double r0, r1, r2;
};

#endif // ELLIPSOID_H
