// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#include "macros.h"
#include <OpenTissue/configuration.h>

#include "pf_skeleton_tracker.h"
#include "auxil.h"
#include "options.h"

#include "cxcore.h"
#include "cv.h"
#include "highgui.h"

const double scale = 2.0;
//const double scale = 1.0;

#define BUMBLEBEE
//#define KINECT

/**
 A basic class for visualising tracking results in the image plane.
 */
template <class tracker_type>
class ProjectedTrackerGUI
{
public:
  typedef typename tracker_type::state_type state_type;
  typedef typename mesh_type::const_face_iterator const_face_iterator;
  
  ProjectedTrackerGUI (const std::string _window, const options &opts, const bool _playback)
    : window (_window),
      left_image_template (opts.left_image_template ()),
      calib (opts),
      track (opts, calib, _playback),
      playback (_playback),
      im (NULL), scaled (NULL),
      frame (opts.data_start_idx ()),
      first_frame (frame),
      object_points (NULL),
      image_points (NULL),
      num_faces (10)
    {
      cvNamedWindow (window.c_str (), CV_WINDOW_AUTOSIZE);
      cvMoveWindow (window.c_str (), 50, 50);
      
      // Get camera matrix (XXX: Read this from 'results.yml' (via 'options'))
      camera_matrix = cvCreateMat (3, 3, CV_64FC1);
      #ifdef BUMBLEBEE
      CV_MAT_ELEM (*camera_matrix, double, 0, 0) = 841.89135742;
      CV_MAT_ELEM (*camera_matrix, double, 0, 1) = 0.0;
      CV_MAT_ELEM (*camera_matrix, double, 0, 2) = 502.90573120;
      CV_MAT_ELEM (*camera_matrix, double, 1, 0) = 0.0;
      CV_MAT_ELEM (*camera_matrix, double, 1, 1) = 843.01513672;
      CV_MAT_ELEM (*camera_matrix, double, 1, 2) = 385.30694580;
      CV_MAT_ELEM (*camera_matrix, double, 2, 0) = 0.0;
      CV_MAT_ELEM (*camera_matrix, double, 2, 1) = 0.0;
      CV_MAT_ELEM (*camera_matrix, double, 2, 2) = 1.0;
      #endif // BUMBLEBEE

      #ifdef KINECT
      CV_MAT_ELEM (*camera_matrix, double, 0, 0) = 465.52684;
      CV_MAT_ELEM (*camera_matrix, double, 0, 1) = 0.0;
      CV_MAT_ELEM (*camera_matrix, double, 0, 2) = 319.50000;
      CV_MAT_ELEM (*camera_matrix, double, 1, 0) = 0.0;
      CV_MAT_ELEM (*camera_matrix, double, 1, 1) = 461.33246;
      CV_MAT_ELEM (*camera_matrix, double, 1, 2) = 228.50000;
      CV_MAT_ELEM (*camera_matrix, double, 2, 0) = 0.0;
      CV_MAT_ELEM (*camera_matrix, double, 2, 1) = 0.0;
      CV_MAT_ELEM (*camera_matrix, double, 2, 2) = 1.0;
      #endif // KINECT

      const matrix3 R = opts.external_rotation ();
      CvMat *rot_mat = cvCreateMat (3, 3, CV_64FC1);
      rot_vec = cvCreateMat (3, 1, CV_64FC1);
      for (int r = 0; r < 3; r++)
        for (int c = 0; c < 3; c++)
          CV_MAT_ELEM (*rot_mat, double, r, c) = R (r, c);
      cvRodrigues2 (rot_mat, rot_vec);
      cvReleaseMat (&rot_mat);

      const vector3 translation = calib.get_world_translation ();
      trans_vec = cvCreateMat (3, 1, CV_64FC1);
      for (int r = 0; r < 3; r++)
        CV_MAT_ELEM (*trans_vec, double, r, 0) = translation (r);
      
      // Invert the rotation and the translation
      for (int r = 0; r < 3; r++)
        {
          CV_MAT_ELEM (*trans_vec, double, r, 0) = -CV_MAT_ELEM (*trans_vec, double, r, 0);
          CV_MAT_ELEM (*rot_vec, double, r, 0) = -CV_MAT_ELEM (*rot_vec, double, r, 0);
        }
    }
  
  ~ProjectedTrackerGUI ()
    {
      if (camera_matrix != NULL)
        cvReleaseMat (&camera_matrix);
      if (rot_vec != NULL)
        cvReleaseMat (&rot_vec);
      if (trans_vec != NULL)
        cvReleaseMat (&trans_vec);
      if (object_points != NULL)
        cvReleaseMat (&object_points);
      if (image_points != NULL)
        cvReleaseMat (&image_points);
    }
  
  void draw_pose (const state_type &pose)
    {
      #ifdef BUMBLEBEE
      draw_cylinder_pose (pose);
      #endif // BUMBLEBEE
      
      #ifdef KINECT
      draw_mesh_pose (pose);
      #endif // KINECT
    }
  
  void draw_cylinder_pose (const state_type &pose)
    {
      if (object_points != NULL)
        cvReleaseMat (&object_points);
      if (image_points != NULL)
        cvReleaseMat (&image_points);      
      object_points = cvCreateMat (4*num_faces, 3, CV_64FC1);
      image_points = cvCreateMat (4*num_faces, 2, CV_64FC1);

      const vector3 root = pose.get_root ();
      const vector3 origin = track.get_calibration ().camera_centre ();

      const double delta_t = M_PI / (double)(num_faces-1);

      //cvFlip (scaled);
      
      for (const_bone_iter iter = pose.bone_begin (); iter != pose.bone_end (); iter++)
        {
          if (iter->is_root ())
            continue;

          int k = 0;

          const double radius = iter->get_width ();
          
          const vector3 a = iter->absolute ().T () + root;
          const vector3 b = iter->parent ()->absolute ().T () + root;
          const vector3 v1 = a - b;
          const vector3 ab2 = b + v1 / 2.0 - origin;
          const vector3 v2 = unit (v1 % ab2); // orthogonal to v1 and ab2
    
          const matrix3 Rt = OpenTissue::math::Ru (delta_t, v1);
          vector3 Rv2 = radius * v2;
          for (double t = 0; t <= M_PI; t += delta_t)
            {
              const vector3 a1 = a + Rv2;
              const vector3 b1 = b + Rv2;

              for (int c = 0; c < 3; c++)
                CV_MAT_ELEM (*object_points, double, k, c) = a1 (c);
              k++;

              for (int c = 0; c < 3; c++)
                CV_MAT_ELEM (*object_points, double, k, c) = b1 (c);
              k++;
          
              Rv2 = Rt * Rv2;
          
              const vector3 b2 = b + Rv2;
              const vector3 a2 = a + Rv2;

              for (int c = 0; c < 3; c++)
                CV_MAT_ELEM (*object_points, double, k, c) = a2 (c);
              k++;

              for (int c = 0; c < 3; c++)
                CV_MAT_ELEM (*object_points, double, k, c) = b2 (c);
              k++;
            }
          
          cvProjectPoints2 (object_points, rot_vec, trans_vec, camera_matrix, NULL, image_points);
          
          // XXX: Can we use 'cvPolyline' instead?
          for (int colour = 0; colour < 2; colour ++) // starting from 0 gives a black boundary if you prefer that
            {
              const CvScalar rgb = CV_RGB (200*colour, 200*colour, 0*colour);
              const int thickness = (colour == 0) ? 3 : 1;
              for (int r = 0; r < image_points->rows; r += 4)
                {
                  CvPoint a1 = cvPoint (CV_MAT_ELEM (*image_points, double, r,   0) / scale,
                                        CV_MAT_ELEM (*image_points, double, r,   1) / scale);
                  CvPoint b1 = cvPoint (CV_MAT_ELEM (*image_points, double, r+1, 0) / scale,
                                        CV_MAT_ELEM (*image_points, double, r+1, 1) / scale);
                  CvPoint a2 = cvPoint (CV_MAT_ELEM (*image_points, double, r+2, 0) / scale,
                                        CV_MAT_ELEM (*image_points, double, r+2, 1) / scale);
                  CvPoint b2 = cvPoint (CV_MAT_ELEM (*image_points, double, r+3, 0) / scale,
                                        CV_MAT_ELEM (*image_points, double, r+3, 1) / scale);
              
                  #define LINE(A, B) cvLine (scaled, A, B, rgb, thickness, CV_AA)
                  if (r == 0 || r == image_points->rows - 4)
                    {
                      LINE (a1, b1);
                      LINE (a2, b2);
                    }
                  LINE (a1, a2);
                  LINE (b1, b2);
                  #undef LINE
                }
            }
        }
        
      //cvFlip (scaled);
    }
  
  bool is_front_facing (const_face_iterator &f, const vector3 &origin)
    {
      vector3 normal;
      OpenTissue::trimesh::compute_face_normal (*f, normal);
      
    
      const vector3 v0 = f->get_vertex0_iterator ()->m_coord;
      const vector3 v1 = f->get_vertex1_iterator ()->m_coord;
      const vector3 v2 = f->get_vertex2_iterator ()->m_coord;
      const vector3 vm = (v0 + v1 + v2) / 3.0 - origin;
      const vector3 vmu = unit (vm);
      
      const double d = dot (normal, vmu);
      
      return (d > 0);
    }
  
  void draw_mesh_pose (const state_type &pose)
    {
      const vector3 root = pose.get_root ();
      const vector3 origin = track.get_calibration ().camera_centre ();
      const mesh_type &mesh = pose.get_mesh ();
      
      int mesh_faces = 0;
      for (const_face_iterator f = mesh.face_begin (); f != mesh.face_end (); f++)
        if (is_front_facing (f, origin))
          mesh_faces ++;
      
      if (object_points != NULL)
        cvReleaseMat (&object_points);
      if (image_points != NULL)
        cvReleaseMat (&image_points);      
      object_points = cvCreateMat (3*mesh_faces, 3, CV_64FC1);
      image_points = cvCreateMat (3*mesh_faces, 2, CV_64FC1);

      cvFlip (scaled);

      int k = 0;
      for (const_face_iterator f = mesh.face_begin (); f != mesh.face_end (); f++)
        {
          if (!is_front_facing (f, origin))
            continue;
            
          const vector3 v0 = f->get_vertex0_iterator ()->m_coord;
          const vector3 v1 = f->get_vertex1_iterator ()->m_coord;
          const vector3 v2 = f->get_vertex2_iterator ()->m_coord;
          
          for (int c = 0; c < 3; c++)
            CV_MAT_ELEM (*object_points, double, k, c) = v0 (c) + root (c);
          k++;

          for (int c = 0; c < 3; c++)
            CV_MAT_ELEM (*object_points, double, k, c) = v1 (c) + root (c);
          k++;

          for (int c = 0; c < 3; c++)
            CV_MAT_ELEM (*object_points, double, k, c) = v2 (c) + root (c);
          k++;
        }
      
      cvProjectPoints2 (object_points, rot_vec, trans_vec, camera_matrix, NULL, image_points);

      #define LINE(A, B) cvLine (scaled, A, B, rgb, thickness, CV_AA)
      for (int colour = 0; colour < 2; colour ++)
        {
          const CvScalar rgb = CV_RGB (200*colour, 200*colour, 0*colour);
          const int thickness = (colour == 0) ? 3 : 1;
          for (int r = 0; r < 3*mesh_faces; r += 3)
            {
              const CvPoint v0 = cvPoint (CV_MAT_ELEM (*image_points, double, r,   0) / scale,
                                          CV_MAT_ELEM (*image_points, double, r,   1) / scale);
              const CvPoint v1 = cvPoint (CV_MAT_ELEM (*image_points, double, r+1, 0) / scale,
                                          CV_MAT_ELEM (*image_points, double, r+1, 1) / scale);
              const CvPoint v2 = cvPoint (CV_MAT_ELEM (*image_points, double, r+2, 0) / scale,
                                          CV_MAT_ELEM (*image_points, double, r+2, 1) / scale);
    
              LINE (v0, v1);
              LINE (v1, v2);
              LINE (v0, v2);
            }
        }
      #undef LINE

      cvFlip (scaled);
    }
  
  void show ()
    {
      const state_type &pose = track.get_pose_estimate ();
      draw_pose (pose);

//      const std::vector<state_type*> &particles = track.get_states ();
//      for (size_t p = 0; p < particles.size (); p++)
//        draw_pose (*(particles [p]));
      
      cvShowImage (window.c_str (), scaled);
    }
    
  void run ()
    {
      while (true)
        {
          const bool valid = refresh ();
          if (!valid)
            break;
            
          show ();
          
          char filename [500];
          sprintf (filename, "screen_%.4d.png", frame);
          cvSaveImage (filename, scaled);

          const int key = cvWaitKey (25);
          if ((char)key == 'q')
            break;
        }
    }

  bool refresh ()
    {
      const bool valid = track.refresh ();
      if (valid)
        {
          track.prepare_mesh ();
          if (!playback)
            track.save_states ();

          if (im != NULL)
            cvReleaseImage (&im);
          char filename [500];
          sprintf (filename, left_image_template.c_str (), frame);
          im = cvLoadImage (filename);
          
          if (scaled != NULL)
            cvReleaseImage (&scaled);
          const CvSize imsize = cvGetSize (im);
          const CvSize ssize = cvSize (imsize.width / scale, imsize.height / scale);
          scaled = cvCreateImage (ssize, IPL_DEPTH_8U, 3);
          cvResize (im, scaled);
          
          frame ++;
        }
      
      return valid;
    }

protected:
  const std::string window, left_image_template;
  calibration calib;
  tracker_type track;
  const bool playback;
  IplImage *im, *scaled;
  int frame, first_frame;
  CvMat *object_points, *rot_vec, *trans_vec, *camera_matrix, *image_points;
  const int num_faces;
};

#define EXPAND(STATE_NAME, STATE_TYPE, OBS_NAME, OBS_TYPE)                              \
        if (st == STATE_NAME && ot == OBS_NAME)                                         \
          {                                                                             \
            ProjectedTrackerGUI< TRACKER (STATE_TYPE, OBS_TYPE) >                       \
                    tracker_gui (#STATE_TYPE " " #OBS_TYPE " Tracker", opts, playback); \
            tracker_gui.run ();                                                         \
          }

int main (int argc, char *argv [])
{
  // Disable localisation as that can screw with punctuation in numerical IO
  setenv ("LANG", "", 1);

  // Command line options
  const std::string options_filename = (argc > 1) ? argv [1] : "";
  const options opts (options_filename);
  const bool playback = (argc > 2 && argv [2] == std::string ("--playback"));

  // Create GUI
  const tracker_state_type st = opts.state_type ();
  const tracker_observation_type ot = opts.observation_type ();

  EXPAND_ALL_TRACKERS

  return 0;
}


