// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef HULL_STATE_H
#define HULL_STATE_H

#include <vector>
#include "angle_state.h"
#include "options.h"
#include "simulator.h"

/**
 An implementation of the projected prior that is constrained by 3D data.
 If you use this class we recommend citing
 @code
 @inproceedings{hauberg:emmcvpr11,
  author    = {S{\o}ren Hauberg and Kim Steenstrup Pedersen},
  booktitle = {Energy Minimization Methods in Computer Vision and Pattern Recognition},
  editor    = {Yuri Boykov and others},
  pages     = {},
  publisher = {Springer},
  series    = {Lecture Notes in Computer Science},
  title     = {{Data-Driven Importance Distributions for Articulated Tracking}},
  year      = {2011}
 }
 @endcode
 */
template<class observation_type>
class hull_state : public skeleton_state<observation_type>
{
public:
  hull_state (const options &opts, const calibration &_calib, const bool _for_show = true)
    : skeleton_state<observation_type> (opts, _calib, _for_show),
      pred_noise (opts.pred_noise ())
    {
      this->compute_pose ();
  
      // Store default chains temporarely
      std::list<vector3> chain_list;
      for (chain_iter iter = this->solver.chain_begin (); iter != this->solver.chain_end (); iter++)
        chain_list.push_back (iter->get_end_effector ()->absolute ().T ());
  
      // Add everything as end-effectors
      int i = 0;
      for (bone_iter iter = this->bone_begin (); iter != this->bone_end (); iter++, i++)
        {
          if (iter->is_root ()) // || i == 3 || i == 4 || i == 5 ||  i == 7 || i == 11)
            continue;
        
          const vector3 absolute = iter->absolute ().T ();
          const vector3 pabsolute = iter->parent ()->absolute ().T ();

          // Make sure the bone isn't already an end-effector
          bool accept = true; //(absolute != pabsolute);
          for (std::list<vector3>::const_iterator ci = chain_list.begin ();
               ci != chain_list.end (); ci ++)
            {
              if (*ci == absolute)
                {
                  accept = false;
                  break;
                }
            }
      
          if (accept)
            {
              chain_type chain;
              chain.init (this->skeleton.root (), &(*iter));
              this->solver.add_chain (chain);
              radius.push_back (iter->get_width ());
            }
        }
    };

  double predict (const observation_type &observation, const double variance_scale = 1.0)
    {
      gaussian1D gauss;
      const knn_cloud &nearest_cloud = observation.get_knn_cloud ();  
      const project2d &proj2d = observation.get_project2d ();

      // Predict root
      //root (0) += 0.05 * gauss.random ();
      //root (1) += 0.0001 * gauss.random ();
      //root (2) += 0.05 * gauss.random ();

      this->compute_pose ();
      vector3 nu, tmp;
      int k = 0;

      const size_t max_rejects = 100;
      size_t chain_idx = 0;
      for (chain_iter iter = this->solver.chain_begin (); iter != this->solver.chain_end (); iter++, chain_idx++)
        {
          const double s = (k >= 7) ? 3.0 * pred_noise :  pred_noise;
          k++;
          const double r = radius [chain_idx];
          
          for (size_t trial = 0; trial < max_rejects; trial++)
            {
              for (size_t n = 0; n < 3; n++)
                tmp [n] = s * gauss.random ();      

              nu = iter->get_end_effector ()->absolute ().T () + tmp; // - this->root;
              
              const CvPoint nu2d = this->calib.world_to_pixel (nu, PROJ2D_SCALE);
              if (proj2d.is_inside (nu2d) && nearest_cloud.nearest_distance (nu) < r)
                break;
            }
          
          iter->p_global () = nu;
        }

      this->solver.solve (&solver_type::default_SD_settings ());
      OpenTissue::kinematics::inverse::get_joint_parameters (*(this->solver.skeleton ()),
                                                             this->theta);

      return 0;
    };
  
  hull_state& operator= (hull_state &new_state)
    {
      skeleton_state<observation_type>::operator= (new_state);

      return *this;
    };
    
  bool load (const std::string filename)
    {
      return skeleton_state<observation_type>::load (filename);
    };

  bool load (std::ifstream &file)
    {
      return skeleton_state<observation_type>::load (file);
    };

  private:
    const double pred_noise;
    std::vector<double> radius;
};

#endif // HULL_STATE_H
