/**
 * This file is part of the nestk library.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Nicolas Burrus <nicolas.burrus@uc3m.es>, (C) 2010
 * Author: Jorge Garcia Bueno <jgarcia@ing.uc3m.es>, (C) 2010
 */

#include "calibration.h"

#include <ntk/ntk.h>
#include <ntk/utils/opencv_utils.h>
#include <ntk/geometry/pose_3d.h>
#include <ntk/mesh/mesh.h>

#include <opencv/highgui.h>

using namespace ntk;
using namespace cv;

namespace ntk
{

RGBDCalibration :: ~RGBDCalibration()
{
  delete depth_pose;
  delete rgb_pose;
}

void RGBDCalibration :: loadFromFile(const char* filename)
{
  //QFileInfo f (filename);
  //ntk_throw_exception_if(!f.exists(), "Could not find calibration file."); // XXX: Do stuff here
  cv::FileStorage calibration_file (filename, CV_STORAGE_READ);
  readMatrix(calibration_file, "rgb_intrinsics", rgb_intrinsics);
  readMatrix(calibration_file, "rgb_distortion", rgb_distortion);
  zero_rgb_distortion = rgb_distortion(0,0) < 1e-5 ? true : false;
  readMatrix(calibration_file, "depth_intrinsics", depth_intrinsics);
  readMatrix(calibration_file, "depth_distortion", depth_distortion);
  zero_depth_distortion = depth_distortion(0,0) < 1e-5 ? true : false;;
  readMatrix(calibration_file, "R", R);
  readMatrix(calibration_file, "T", T);
  cv::Mat1i size_mat;
  readMatrix(calibration_file, "rgb_size", size_mat);
  rgb_size = cv::Size(size_mat(0,0), size_mat(0,1));
  readMatrix(calibration_file, "raw_rgb_size", size_mat);
  raw_rgb_size = cv::Size(size_mat(0,0), size_mat(0,1));
  cv::Mat1i rgb_size_mat;
  readMatrix(calibration_file, "depth_size", size_mat);
  depth_size = cv::Size(size_mat(0,0), size_mat(0,1));
  readMatrix(calibration_file, "raw_depth_size", size_mat);
  raw_depth_size = cv::Size(size_mat(0,0), size_mat(0,1));

  cv::Mat1f depth_calib (1,2);
  try {
    readMatrix(calibration_file, "depth_base_and_offset", depth_calib);
    depth_baseline = depth_calib(0,0);
    depth_offset = depth_calib(0,1);
  }
  catch(...)
  {
  }

  calibration_file.release();

  depth_pose = new Pose3D();
  depth_pose->setCameraParametersFromOpencv(depth_intrinsics);

  rgb_pose = new Pose3D();
  rgb_pose->toRightCamera(rgb_intrinsics, R, T);

  initUndistortRectifyMap(rgb_intrinsics, rgb_distortion,
                          Mat(), rgb_intrinsics, rgb_size, CV_16SC2,
                          rgb_undistort_map1, rgb_undistort_map2);

  initUndistortRectifyMap(depth_intrinsics, depth_distortion,
                          Mat(), depth_intrinsics, depth_size, CV_16SC2,
                          depth_undistort_map1, depth_undistort_map2);
}

} // ntk

