/**
 * This file is part of the nestk library.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Author: Nicolas Burrus <nicolas.burrus@uc3m.es>, (C) 2010
 */

#include "mesh_generator.h"

#include <ntk/ntk.h>
#include <ntk/geometry/pose_3d.h>

using namespace cv;

#ifdef _MSC_VER
	inline int rnd(float  x) { return floor(x + 0.5f); }
    inline int rnd(double x) { return floor(x + 0.5); }
#else
	inline int rnd(float  x) { return lroundf(x); }
    inline int rnd(double x) { return lround(x); }
#endif 

namespace ntk
{

  MeshGenerator :: MeshGenerator()
    : m_use_color(false),
    m_mesh_type(PointCloudMesh),
    m_resolution_factor(1.0)
  {
  }

  void MeshGenerator :: setResolutionFactor(double f)
  {
    m_resolution_factor = f;
  }

  void MeshGenerator :: generatePointCloudMesh(const RGBDImage& image,
                                               const Pose3D& depth_pose,
                                               const Pose3D& rgb_pose)
  {
    m_mesh.clear();
    m_mesh.vertices.reserve(image.depth().rows*image.depth().cols);
    m_mesh.colors.reserve(image.depth().rows*image.depth().cols);
    m_mesh.normals.reserve(image.depth().rows*image.depth().cols);

    const cv::Mat1f& depth_im = image.depth();
    const cv::Mat1b& mask_im = image.depthMask();
    cv::Mat3f voxels (depth_im.size());
    cv::Mat3f rgb_points (depth_im.size());

    cv::Mat1b subsample_mask(mask_im.size());
    subsample_mask = 0;
    for (float r = 0; r < subsample_mask.rows-1; r += 1.0/m_resolution_factor)
      for (float c = 0; c < subsample_mask.cols-1; c += 1.0/m_resolution_factor)
        subsample_mask(rnd(r), rnd(c)) = 1;
    subsample_mask = mask_im & subsample_mask;

    depth_pose.unprojectFromImage(depth_im, subsample_mask, voxels);
    if (m_use_color)
      rgb_pose.projectToImage(voxels, subsample_mask, rgb_points);

    for (int r = 0; r < voxels.rows; ++r)
    {
      Vec3f* voxels_data = voxels.ptr<Vec3f>(r);
      //const uchar* mask_data = subsample_mask.ptr<uchar>(r); // XXX: Re-enable??
      for (int c = 0; c < voxels.cols; ++c)
      {
        // XXX: Re-enable??
        //if (!mask_data[c])
        //  continue;

        bool works = false;
        Vec3b color (0,0,0);
        if (m_use_color)
        {
          Point3f prgb = rgb_points(r,c);
          int i_y = rnd(prgb.y);
          int i_x = rnd(prgb.x);
          if (is_yx_in_range(image.rgb(), i_y, i_x))
          {
            works = true;
            Vec3b bgr = image.rgb()(i_y, i_x);
            color = Vec3b(bgr[2], bgr[1], bgr[0]);
          }
        }
        else
        {
          int g = 0;
          if (image.intensity().data)
            g = image.intensity()(r,c);
          else
            g = 255 * voxels_data[c][2] / 10.0;
          color = Vec3b(g,g,g);
        }

        m_mesh.vertices.push_back(voxels_data[c]);
        m_mesh.colors.push_back(color);
      }
    }
  }

  void MeshGenerator :: generate(const RGBDImage& image,
                                 const Pose3D& depth_pose,
                                 const Pose3D& rgb_pose)
  {;
    const Pose3D& real_depth_pose = depth_pose.isValid() ? depth_pose : *(image.calibration()->depth_pose);
    const Pose3D& real_rgb_pose = rgb_pose.isValid() ? rgb_pose : *(image.calibration()->rgb_pose);

    return generatePointCloudMesh(image, real_depth_pose, real_rgb_pose);
  }

} // ntk
