// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#include "options.h"
#include <iostream>
#include <fstream>

// Constructor
options::options (const std::string filename)
  : valid_data_template (false),
    valid_data_start_idx (false),
    valid_output_directory (false),
    valid_state_type (false),
    valid_num_particles (false),
    valid_initial_pose (false),
    valid_pred_noise (false),
    val_extrapolate (false),
    valid_convex_lambda (false),
    valid_intrinsic_matrix (false),
    val_data_scaling (1.0),
    valid_external_view (false),
    valid_external_translation (false),
    valid_external_rotation (false),
    valid_kinect_calibration (false),
    valid_depth_background (false),
    valid_skeleton_xsf (false),
    valid_bone_appearance (false),
    valid_joint_limits (false),
    valid_bumblebee_configuration (false),
    valid_observation_type (false),
    valid_measure_threshold (false),
    valid_measure_variance (false),
    valid_depth_threshold (false),
    val_measure_downsampling (1),
    valid_stick_filename (false),
    valid_mocap_filename (false)
{
  // Read general-purpose options file
  std::ifstream file (filename.c_str ());
  if (!file.is_open ())
    {
      std::cerr << "options::options: couldn't open '"
                << filename << "' for reading" << std::endl;
      return;
    }
    
  while (!file.eof ())
    {
      char rest_of_line [100];
      std::string option_name;
      file >> option_name;
      if (option_name [0] == '#')
        {
          file.getline (rest_of_line, sizeof (rest_of_line));
          continue;
        }
      
      if (option_name == "data_template")
        {
          file >> val_data_template;
          valid_data_template = true;
        }
      else if (option_name == "data_start_idx")
        {
          file >> val_data_start_idx;
          valid_data_start_idx = true;
        }
      else if (option_name == "left_image_template")
        {
          file >> val_left_image_template;
          valid_left_image_template = true;
        }
      else if (option_name == "right_image_template")
        {
          file >> val_right_image_template;
          valid_right_image_template = true;
        }
      else if (option_name == "background_image")
        {
          file >> val_background_image;
          valid_background_image = true;
        }
      else if (option_name == "output_directory")
        {
          file >> val_output_directory;
          valid_output_directory = true;
        }
      else if (option_name == "state_type")
        {
          std::string str_state_type;
          file >> str_state_type;

          valid_state_type = true;
          if (str_state_type == "angle_state")
            {
              val_state_type = ANGLE_STATE;
            }
          else if (str_state_type == "pik_state")
            {
              val_state_type = PIK_STATE;
            }
          else if (str_state_type == "projection_state")
            {
              val_state_type = PROJECTION_STATE;
            }
          else if (str_state_type == "tangent_state")
            {
              val_state_type = TANGENT_STATE;
            }
          else if (str_state_type == "kbk_stick_state")
            {
              val_state_type = KBK_STICK_STATE;
            }
          else if (str_state_type == "stick_state")
            {
              val_state_type = STICK_STATE;
            }
          else if (str_state_type == "stick2d_state")
            {
              val_state_type = STICK2D_STATE;
            }
          else if (str_state_type == "brownian_stick_state")
            {
              val_state_type = BROWNIAN_STICK_STATE;
            }
          else if (str_state_type == "pik_stick_state")
            {
              val_state_type = PIK_STICK_STATE;
            }
          else if (str_state_type == "brownian_state")
            {
              val_state_type = BROWNIAN_STATE;
            }
          else if (str_state_type == "bgsub_state")
            {
              val_state_type = BGSUB_STATE;
            }
          else if (str_state_type == "angle_bgsub_state")
            {
              val_state_type = ANGLE_BGSUB_STATE;
            }
          else if (str_state_type == "hull_state")
            {
              val_state_type = HULL_STATE;
            }
          else
            {
              std::cerr << "Uknown state type '" << str_state_type << "'" << std::endl;
              valid_state_type = false;
            }
        }
      else if (option_name == "num_particles")
        {
          file >> val_num_particles;
          valid_num_particles = true;
        }
      else if (option_name == "initial_pose")
        {
          file >> val_initial_pose;
          valid_initial_pose = true;
        }
      else if (option_name == "pred_noise")
        {
          file >> val_pred_noise;
          valid_pred_noise = true;
        }
      else if (option_name == "extrapolate")
        {
          file >> val_extrapolate;
        }
      else if (option_name == "convex_lambda")
        {
          file >> val_convex_lambda;
          valid_convex_lambda = true;
        }
      else if (option_name == "camera_calibration")
        {
          std::string option_value;
          file >> option_value;
          //read_camera_calibration (option_value);
        }
      else if (option_name == "kinect_calibration")
        {
          file >> val_kinect_calibration;
          valid_kinect_calibration = true;
        }
      else if (option_name == "depth_background")
        {
          file >> val_depth_background;
          valid_depth_background = true;
        }
      else if (option_name == "external_view")
        {
          file >> val_external_view;
          valid_external_view = true;
          
          valid_external_translation = valid_external_rotation
            = read_external_view (val_external_view);
        }
      else if (option_name == "intrinsic_matrix")
        {
          double fx, fy, cx, cy;
          file >> fx;
          file >> fy;
          file >> cx;
          file >> cy;
          val_intrinsic_matrix (0, 0) = fx;
          val_intrinsic_matrix (1, 1) = fy;
          val_intrinsic_matrix (0, 2) = cx;
          val_intrinsic_matrix (1, 2) = cy;
          val_intrinsic_matrix (2, 2) = 1.0;
          
          valid_intrinsic_matrix = true;          
        }
      else if (option_name == "data_scaling")
        {
          file >> val_data_scaling;
        }
      else if (option_name == "bone_appearance")
        {
          file >> val_bone_appearance;
          valid_bone_appearance = true;
          read_bone_appearance (val_bone_appearance);
        }
      else if (option_name == "joint_limits")
        {
          file >> val_joint_limits;
          valid_joint_limits = true;
        }
      else if (option_name == "skeleton_xsf")
        {
          file >> val_skeleton_xsf;
          valid_skeleton_xsf = true;
        }
      else if (option_name == "bumblebee_configuration")
        {
          file >> val_bumblebee_configuration;
          valid_bumblebee_configuration = true;
        }
      else if (option_name == "observation_type")
        {
          std::string str_observation_type;
          file >> str_observation_type;
          
          valid_observation_type = true;
          if (str_observation_type == "point_cloud")
            {
              val_observation_type = POINT_CLOUD;
            }
          else if (str_observation_type == "point_cloud_and_stick")
            {
              val_observation_type = POINT_CLOUD_AND_STICK;
            }
          else if (str_observation_type == "kinect_cloud")
            {
              val_observation_type = KINECT_CLOUD;
            }
          else if (str_observation_type == "triclops_cloud")
            {
              val_observation_type = TRICLOPS_CLOUD;
            }
          else
            {
              std::cerr << "Unknown observation type '" << str_observation_type
                        << "'" << std::endl;
              valid_observation_type = false;
            }
        }
      else if (option_name == "measure_threshold")
        {
          file >> val_measure_threshold;
          valid_measure_threshold = true;
        }
      else if (option_name == "measure_variance")
        {
          file >> val_measure_variance;
          valid_measure_variance = true;
        }
      else if (option_name == "depth_threshold")
        {
          file >> val_depth_threshold;
          valid_depth_threshold = true;
        }
      else if (option_name == "measure_downsampling")
        {
          file >> val_measure_downsampling;
        }
      else if (option_name == "stick_filename")
        {
          file >> val_stick_filename;
          valid_stick_filename = true;
        }
      else if (option_name == "mocap_filename")
        {
          file >> val_mocap_filename;
          valid_mocap_filename = true;
        }
      else if (!option_name.empty ())
        {
          std::string option_value;
          file >> option_value;
          std::cerr << "options::options: unknown configuration: "
                    << option_name << " " << option_value << std::endl;
        }

      file.getline (rest_of_line, sizeof (rest_of_line));  
    }
  file.close ();
}

// Input data
std::string options::data_template () const
{
  if (!valid_data_template)
    throw missing_option ("Mising option: data_template");
  
  return val_data_template;
}

int options::data_start_idx () const
{
  if (!valid_data_start_idx)
    throw missing_option ("Mising option: data_start_idx");
  
  return val_data_start_idx;
}

std::string options::left_image_template () const
{
  if (!valid_left_image_template)
    throw missing_option ("Missing option: left_image_template");
  
  return val_left_image_template;
}

std::string options::right_image_template () const
{
  if (!valid_right_image_template)
    throw missing_option ("Missing option: right_image_template");
  
  return val_right_image_template;
}

std::string options::background_image () const
{
  if (!valid_background_image)
    throw missing_option ("Missing option: background_image");
  
  return val_background_image;
}

// Output data
std::string options::output_directory () const
{
  if (!valid_output_directory)
    throw missing_option ("Mising option: output_directory");
  
  return val_output_directory;
}

// Tracker options
tracker_state_type options::state_type () const
{
  if (!valid_state_type)
    throw missing_option ("Missing option: state_type");
  
  return val_state_type;
}

int options::num_particles () const
{
  if (!valid_num_particles)
    throw missing_option ("Mising option: num_particles");
  
  return val_num_particles;
}

std::string options::initial_pose () const
{
  if (!valid_initial_pose)
    throw missing_option ("Mising option: initial_pose");
  
  return val_initial_pose;
}

double options::pred_noise () const
{
  if (!valid_pred_noise)
    throw missing_option ("Mising option: pred_noise");
  
  return val_pred_noise;
}

bool options::extrapolate () const
{
  return val_extrapolate;
}

double options::convex_lambda () const
{
  if (!valid_convex_lambda)
    throw missing_option ("Mising option: convex_lambda");
  
  return val_convex_lambda;
}

void options::read_bone_appearance (const std::string filename)
{
  std::ifstream file (filename.c_str ());
  if (!file.is_open ())
    {
      std::cerr << "options::read_bone_appearance: couldn't open '"
                << filename << "' for reading" << std::endl;
      return;
    }
    
  while (!file.eof ())
    {
      std::string bone_name;
      double bone_width;
      file >> bone_name;
      if (bone_name [0] == '#')
        continue;
      
      file >> bone_width;
      val_bone_size [bone_name] = bone_width;
    }
  file.close ();
}

std::string options::external_view () const
{
  if (!valid_external_view)
    throw missing_option ("Missing option: external_view");
  
  return val_external_view;
}

const vector3& options::external_translation () const
{
  if (!valid_external_translation)
    throw missing_option ("Missing option: external_translation");
  
  return val_external_translation;
}

const matrix3& options::external_rotation () const
{
  if (!valid_external_rotation)
    throw missing_option ("Missing option: external_rotation");
  
  return val_external_rotation;
}

bool options::read_external_view (const std::string filename)
{
  bool retval = false;
  std::ifstream rtfile (filename.c_str ());
  if (rtfile.is_open ())
    {
      rtfile >> val_external_translation (0);
      rtfile >> val_external_translation (1);
      rtfile >> val_external_translation (2);
      
      double rot_x, rot_y, rot_z;
      rtfile >> rot_x;
      rtfile >> rot_y;
      rtfile >> rot_z;
      
      const matrix3 rxm = OpenTissue::math::Rx (rot_x);
      const matrix3 rym = OpenTissue::math::Ry (rot_y);
      const matrix3 rzm = OpenTissue::math::Rz (rot_z);
  
      val_external_rotation = rxm * rym * rzm;

      rtfile.close ();
      retval = true;
    }
  else
    std::cerr << "Could not open '" << filename << "' for reading" << std::endl;
  
  return retval;
}

const std::string& options::kinect_calibration () const
{
  if (!valid_kinect_calibration)
    throw missing_option ("Missing option: kinect_calibration");
  
  return val_kinect_calibration;
}

const std::string& options::depth_background () const
{
  if (!valid_depth_background)
    throw missing_option ("Missing option: depth_background");
  
  return val_depth_background;
}

const matrix3& options::intrinsic_matrix () const
{
  if (!valid_intrinsic_matrix)
    throw missing_option ("Missing option: intrinsic_matrix");
  
  return val_intrinsic_matrix;
}

double options::data_scaling () const
{
  return val_data_scaling;
}

// Skeleton definition
std::string options::skeleton_xsf () const
{
  if (!valid_skeleton_xsf)
    throw missing_option ("Mising option: skeleton_xsf");
  
  return val_skeleton_xsf;
}

// Bone appearance
std::string options::bone_appearance () const
{
  if (!valid_bone_appearance)
    throw missing_option ("Mising option: bone_appearance");
  
  return val_bone_appearance;
}

double options::bone_size (const std::string name) const
{
  std::map<const std::string, double>::const_iterator iter = val_bone_size.find (name);
  if (iter == val_bone_size.end ())
    throw missing_option ("Missing option: " + name);
  
  return iter->second;
}

std::string options::joint_limits () const
{
  if (!valid_bone_appearance)
    throw missing_option ("Mising option: joint_limits");
  
  return val_joint_limits;
}
  
// Bumblebee
std::string options::bumblebee_configuration () const
{
  if (!valid_bumblebee_configuration)
    throw missing_option ("Missing option: bumblebee_configuration");
  
  return val_bumblebee_configuration;
}

// Measurement
tracker_observation_type options::observation_type () const
{
  if (!valid_observation_type)
    {
      std::cerr << "Observation type has not been given in config file; defaulting to point_cloud"
                << std::endl;
      return POINT_CLOUD;
    }
  else
    return val_observation_type;
}

double options::measure_threshold () const
{
  if (!valid_measure_threshold)
    throw missing_option ("Missing option: measure_threshold");
  
  return val_measure_threshold;
}

double options::measure_variance () const
{
  if (!valid_measure_variance)
    throw missing_option ("Missing option: measure_variance");
  
  return val_measure_variance;
}

double options::depth_threshold () const
{
  if (!valid_depth_threshold)
    throw missing_option ("Missing option: depth_threshold");
  
  return val_depth_threshold;
}

int options::measure_downsampling () const
{
  return val_measure_downsampling;
}

// Stick data
std::string options::stick_filename () const
{
  if (!valid_stick_filename)
    throw missing_option ("Missing option: stick_filename");
  
  return val_stick_filename;
}

// MoCap validation data
std::string options::mocap_filename () const
{
  if (!valid_mocap_filename)
    throw missing_option ("Missing option: mocap_filename");
  
  return val_mocap_filename;
}

