// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef OPTIONS_H
#define OPTIONS_H

#include <stdexcept>
#include <string>
#include <map>
#include "math_types.h"

/**
 Different kinds of states for the particle filter. The difference in states mainly
 consists of the choice of prediction algorithms.
 */
enum tracker_state_type
{
  ANGLE_STATE,
  PIK_STATE,
  PROJECTION_STATE,
  TANGENT_STATE,
  KBK_STICK_STATE,
  STICK_STATE,
  STICK2D_STATE,
  BROWNIAN_STICK_STATE,
  PIK_STICK_STATE,
  BROWNIAN_STATE,
  BGSUB_STATE,
  ANGLE_BGSUB_STATE,
  HULL_STATE
};

/**
 Different kinds of observations.
 */
enum tracker_observation_type
{
  POINT_CLOUD,
  POINT_CLOUD_AND_STICK,
  KINECT_CLOUD,
  TRICLOPS_CLOUD
};

/**
 Exception class for missing options. This is thrown when the configuration file
 is missing the option containing the information needed for a function.
 */
class missing_option : public std::runtime_error
{
public:
  missing_option (const std::string &message = "")
    : std::runtime_error (message)
  {
  }
};

/**
 The options class contains a basic interface for accessing information stored
 in configuration files.
 
 A configuration file as written in plain text with one line per option. The
 first word of the line is the name of the option, while the rest of the line
 should contain the value of the option. As an example, the following text could
 be the contents of a configuration file.
 @code
 observation_type          point_cloud_and_stick
 state_type                stick_state
 data_template             Movies3D/triangles/stick2_060110/stick2_060110_%.3d.points
 @endcode
 In this case the file contains three options, all having string values.
 */
class options
{
public:
  /**
   Constructor for the options class. This class contains the information stored
   in configuration files. The information is read from the configuration file
   once by the constructor and the file is never touched again.
   
   In general function names correspond to the name of the options in the
   configuration file.
   
   @param[in] filename
     The name of the configuration file.
   */
  options (const std::string filename);
  
  // Input data
  /**
   A string in printf format with the path of the file containing the data to be
   analysed. The string should allow for one integer to be insterted by printf.
   This integer will correspond to the number of the current frame.
   
   As an example, the following
   @code
   Movies3D/triangles/stick2_060110/stick2_060110_%.3d.points
   @endcode
   is a valid string for this option.
   */
  std::string data_template () const;
  int data_start_idx () const;
  std::string left_image_template () const;
  std::string right_image_template () const;
  std::string background_image () const;
  
  // Output data
  std::string output_directory () const;
  
  // Tracker options
  tracker_state_type state_type () const;
  int num_particles () const;
  std::string initial_pose () const;
  double pred_noise () const;
  bool extrapolate () const;
  double convex_lambda () const;
  
  // Camera calibration
  const matrix3& intrinsic_matrix () const;
  double data_scaling () const;
  
  std::string external_view () const;
  const vector3& external_translation () const;
  const matrix3& external_rotation () const;
  
  const std::string& kinect_calibration () const;
  const std::string& depth_background () const;
  
  std::string skeleton_xsf () const;
  
  // Bone appearance
  std::string bone_appearance () const;
  double bone_size (const std::string name) const;
  
  std::string joint_limits () const;
  
  // Bumblebee
  std::string bumblebee_configuration () const;
  
  // Measurement
  tracker_observation_type observation_type () const;
  double measure_threshold () const;
  double measure_variance () const;
  double depth_threshold () const;
  int measure_downsampling () const;
  
  // Stick info
  std::string stick_filename () const;
  
  // MoCap validation data
  std::string mocap_filename () const;

private:
  // Input data template
  std::string val_data_template;
  bool valid_data_template;
  int val_data_start_idx;
  bool valid_data_start_idx;
  std::string val_left_image_template;
  bool valid_left_image_template;
  std::string val_right_image_template;
  bool valid_right_image_template;
  std::string val_background_image;
  bool valid_background_image;
  
  // Output data
  std::string val_output_directory;
  bool valid_output_directory;
  
  // Tracker options
  tracker_state_type val_state_type;
  bool valid_state_type;
  int val_num_particles;
  bool valid_num_particles;
  std::string val_initial_pose;
  bool valid_initial_pose;
  double val_pred_noise;
  bool valid_pred_noise;
  bool val_extrapolate;
  bool valid_convex_lambda;
  double val_convex_lambda;
  
  // Camera calibration
  std::string val_external_view;
  vector3 val_external_translation;
  matrix3 val_external_rotation;
  matrix3 val_intrinsic_matrix;
  bool valid_intrinsic_matrix;
  double val_data_scaling;
  
  bool valid_external_view;
  bool valid_external_translation;
  bool valid_external_rotation;
  
  // Kinect stuff
  std::string val_kinect_calibration;
  bool valid_kinect_calibration;
  std::string val_depth_background;
  bool valid_depth_background;

  // Skeleton definition
  std::string val_skeleton_xsf;
  bool valid_skeleton_xsf;

  // Bone appearance
  std::string val_bone_appearance;
  bool valid_bone_appearance;
  std::map<const std::string, double> val_bone_size;
  
  bool valid_joint_limits;
  std::string val_joint_limits;
  
  // Bumblebee
  std::string val_bumblebee_configuration;
  bool valid_bumblebee_configuration;
  
  // Measurement
  tracker_observation_type val_observation_type;
  double val_measure_threshold;
  double val_measure_variance;
  double val_depth_threshold;
  bool valid_observation_type;
  bool valid_measure_threshold;
  bool valid_measure_variance;
  bool valid_depth_threshold;
  int val_measure_downsampling;
  
  // Stick info
  std::string val_stick_filename;
  bool valid_stick_filename;
  
  // MoCap validation data
  std::string val_mocap_filename;
  bool valid_mocap_filename;
  
  // Helper functions
  void read_camera_calibration (const std::string filename);
  void read_bone_appearance (const std::string filename);
  bool read_external_view (const std::string filename);
};

#endif // OPTIONS_H
