// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#include <OpenTissue/configuration.h>

#define DEFINE_GLUT_MAIN
#include <OpenTissue/utility/glut/glut_perspective_view_application.h>
#undef DEFINE_GLUT_MAIN
#include <OpenTissue/utility/gl/gl_draw_ellipsoid.h>
#include <OpenTissue/utility/gl/gl_draw_string.h>

#include "pf_skeleton_tracker.h"
#include "auxil.h"
#include "options.h"
#include "macros.h"
#include "ellipsoid.h"

#include <list>

/**
 A small demo application for visualising the covariance of different priors.
 */
class Application : public OpenTissue::glut::PerspectiveViewApplication
{
protected:
  calibration calib;
  angle_state<point_cloud_and_stick> angle_pose, mean_pose;
  projection_state<point_cloud_and_stick> proj_pose, mean_proj_pose;
  KBK_stick_state<point_cloud_and_stick> kbk_pose, mean_kbk_pose;
  stick_state<point_cloud_and_stick> stick_pose, mean_stick_pose;
  point_cloud_and_stick observation;
  std::vector<ellipsoid*> E_angle, E_kbk, E_proj, E_stick;
  unsigned char view_mode;
  
public:
  Application (const options &opts)
    : calib (opts),
      angle_pose (opts, calib),
      mean_pose (opts, calib),
      proj_pose (opts, calib),
      mean_proj_pose (opts, calib),
      kbk_pose (opts, calib),
      mean_kbk_pose (opts, calib),
      stick_pose (opts, calib),
      mean_stick_pose (opts, calib),
      observation (opts, calib),
      view_mode ('1')
    {
      this->z_far () = 5000; // what does this do?
      read_ellipsoids ("data/ellipses_angle.txt", E_angle);
      read_ellipsoids ("data/ellipses_kbk.txt", E_kbk);
      read_ellipsoids ("data/ellipses_proj.txt", E_proj);
      read_ellipsoids ("data/ellipses_stick.txt", E_stick);
    }

  void read_ellipsoids (const char *filename, std::vector<ellipsoid*> &EV)
    {
      vector3 C, A0, A1, A2;
      double r0, r1, r2;
      std::ifstream file (filename);
      for (int k = 0; k < 14; k++)
        {
          for (int l = 0; l < 3; l++)
            file >> C (l);
          for (int l = 0; l < 3; l++)
            file >> A0 (l);
          for (int l = 0; l < 3; l++)
            file >> A1 (l);
          for (int l = 0; l < 3; l++)
            file >> A2 (l);
          file >> r0;
          file >> r1;
          file >> r2;
          
          r0 = std::max (2.5 * sqrt (r0), 0.01);
          r1 = std::max (2.5 * sqrt (r1), 0.01);
          r2 = std::max (2.5 * sqrt (r2), 0.01);

          EV.push_back (new ellipsoid (C, A0, A1, A2, r0, r1, r2));
        }
      file.close ();
    }

  void show_ellipsoids (std::vector<ellipsoid*> &EV)
    {
      for (size_t k = 0; k < EV.size (); k++)
        {
          OpenTissue::gl::ColorPicker (0.9, 0.1, 0.9, 0.7);
          OpenTissue::gl::DrawEllipsoid<ellipsoid> (*(EV [k]), true);
        }
    }
    
  const char* do_get_title () const
    {
      return "Playground";
    }

  void draw_stick ()
    {
      const stick_type stick = observation.get_stick ();
      OpenTissue::gl::ColorPicker (0.9, 0.1, 0.9, 1.0);
      OpenTissue::gl::DrawVector (stick.second, stick.first - stick.second, 1.0, false);
      OpenTissue::gl::DrawPoint (stick.first, 0.2);
      OpenTissue::gl::DrawPoint (stick.second, 0.2);    
    }
    
  void do_display ()
    {
      static bool in_progress = false;
      if (in_progress) return;
      
      in_progress = true;
      mean_pose.show ();
      
      std::string text ("View Mode =  ");
      text [text.size ()-1] = view_mode;
      OpenTissue::gl::DrawString (text, vector3 (0, -4, 0));
      
      if (view_mode == '1')
        {
          // Show nothing extra
        }
      else if (view_mode == '2')
        {
          angle_pose.show ();
        }
      else if (view_mode == '3')
        {
          // Show angular covariances
          show_ellipsoids (E_angle);
        }
      else if (view_mode == '4')
        {
          kbk_pose.show ();
          draw_stick ();
        }
      else if (view_mode == '5')
        {
          // Show KBK covariances
          show_ellipsoids (E_kbk);
          draw_stick ();
        }
      else if (view_mode == '6')
        {
          proj_pose.show ();
        }
      else if (view_mode == '7')
        {
          // Show projected covariances
          show_ellipsoids (E_proj);
        }
      else if (view_mode == '8')
        {
          stick_pose.show ();
          draw_stick ();
        }
      else if (view_mode == '9')
        {
          // Show stick covariances
          show_ellipsoids (E_stick);
          draw_stick ();
        }

      in_progress = false;
    }
  
  void do_action (unsigned char choice)
    {
      switch (choice)
        {
          case '1':
            view_mode = choice;
            break;
          case '2':
            angle_pose = mean_pose;
            angle_pose.predict (observation);
            view_mode = choice;
            break;
          case '3':
            view_mode = choice;
            break;
          case '4':
            kbk_pose = mean_kbk_pose;
            kbk_pose.predict (observation);
            view_mode = choice;
            break;
          case '5':
            view_mode = choice;
            break;
          case '6':
            proj_pose = mean_proj_pose;
            proj_pose.predict (observation);
            view_mode = choice;
            break;        
          case '7':
            view_mode = choice;
            break;        
          case '8':
            stick_pose = mean_stick_pose;
            stick_pose.predict (observation);
            view_mode = choice;
            break;
          case '9':
            view_mode = choice;
            break;
          default:
            std::cout << "You pressed " << choice << std::endl;
            break;
        };
    }

  void do_init_right_click_menu (int main_menu, void menu (int entry))
    {
    }

  void do_init ()
    {
      this->camera ().move (80);
      glClearColor (0.9, 0.9, 0.9, 0.0);
      glEnable (GL_BLEND);
      glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);  

      std::cout << "Press 't' to get going" << std::endl;
    }

  void do_run ()
    {
    }

  void do_shutdown ()
    {
    }
};

OpenTissue::glut::instance_pointer init_glut_application (int argc, char *argv [])
{
  // Disable localisation as that screws with punctuation in numerical IO
  setenv ("LANG", "", 1);

  // Get options
  const std::string options_filename = (argc > 1) ? argv [1] : "";
  const options opts (options_filename);

  // Create GUI
  OpenTissue::glut::instance_pointer instance;
  instance.reset (new Application (opts));
        
  return instance;
}

