// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef POINT_CLOUD_H
#define POINT_CLOUD_H

#include <string>
#include <vector>

#include "cloud.h"
#include "options.h"
#include "math_types.h"
#include <boost/filesystem.hpp>

/**
 The point_cloud class gives access to point data generated by a stereo camera.
 Specifically it provides access to 3D points.
 */
class point_cloud : public cloud3
{
  public:
    /**
     Constructor for the point_cloud class. The constructor initialises some
     variables, but does not actually read any data. The 'refresh' function is
     used to actually read data.
     
     @param[in] opts
       An object of class options that contains configuration information. The
       point_cloud constructor reads the following options:
       <ul>
         <li>data_template</li>
         <li>data_start_idx</li>
         <li>external_view</li>
         <li>data_scaling (optional)</li>
         <li>depth_threshold (optional)</li>
       </ul>
     
     @param[in] _calib
       A calibration object describing the used coordinate systems.
     */
    point_cloud (const options &opts, const calibration &_calib)
      : cloud3 (opts, _calib),
        input_template (opts.data_template ())
      {
        init ();
      };
    
    virtual bool read_background (const std::string &depth_background)
      {
        point_vector P;
        const int num_frames = 5;
        P.reserve (num_frames * 320 * 240);
        for (int bg_frame = 15; bg_frame < 15+num_frames; bg_frame++)
          {
            const char *format = "%s/points_%d.txt";
            char buffer [100];
            sprintf (buffer, format, depth_background.c_str (), bg_frame);
                
            if (boost::filesystem::is_regular (buffer))
              {
                FILE *fid = fopen (buffer, "r");
                const bool found = (fid != NULL);
                if (!found)
                  {
                    std::cerr << "point_cloud::point_cloud: could not open depth background '"
                              << buffer << "' for reading" << std::endl;
                  }
                else
                  {
                    // Count number of lines in file
                    const int buf_size = 100;
                    char line [buf_size];
                    size_t num_points = 0;
                    while (fgets (line, buf_size, fid) != NULL)
                      num_points ++;
  
                    // Read the data
                    rewind (fid);
                    for (size_t k = 0; k < num_points; k++)
                      {
                        double x, y, z;
                        double r, g, b;
                        const int num_read = fscanf (fid, "%lf %lf %lf %lf %lf %lf\n", &x, &y, &z, &r, &g, &b);
                        if (num_read != 6)
                          {
                            std::cerr << "point_cloud::point_cloud: invalid data format" << std::endl;
                            break;
                          }
                
                        const vector3 point = calib.camera_to_world (vector3 (x, y, z));
                        P.push_back (point);
                      }
                  }

                fclose (fid);
              }
          }

        if (P.size () > 0)
          bg_nearest_cloud.refresh (P);
        
        return (P.size () > 0);
      }

    /**
     Reads data for the next frame. You should call this function when you are ready
     to move on to the next frame.
     
     @param[in] goto_next
       This optional parameter determines if the next frame should be read (default)
       or the current frame should be re-read. The latter is useful if you have
       changed parameters of the point_cloud object, such as the translation or
       rotation of the data.
     
     @return
       If the frame was succesfully read the function returns true, otherwise it
       returns false.
     */
    bool refresh (bool goto_next = true)
      {
        // Read data from disk
        char filename [100];
        sprintf (filename, input_template.c_str (), frame);
        if (goto_next) frame ++;
        
        FILE *fid = fopen (filename, "r");
        const bool found = (fid != NULL);
        if (!found)
          {
            std::cerr << "point_cloud::refresh: could not open '"
                      << filename << "' for reading" << std::endl;
          }
        else
          {
            // Count number of lines in file
            const int buf_size = 100;
            char line [buf_size];
            size_t num_points = 0;
            while (fgets (line, buf_size, fid) != NULL)
              num_points ++;

            // Allocate data storage
            points.clear ();
            points.reserve (num_points / downsample);
            colours.clear ();
            colours.reserve (num_points / downsample);
  
            // Read the data
            rewind (fid);
            const double y_thresh = -0.5; // For hands2
            for (size_t k = 0; k < num_points; k++)
              {
                double x, y, z;
                double r, g, b;
                const int num_read = fscanf (fid, "%lf %lf %lf %lf %lf %lf\n", &x, &y, &z, &r, &g, &b);
                if (num_read != 6)
                  {
                    std::cerr << "point_cloud::refresh: invalid data format" << std::endl;
                    break;
                  }
                
                // Should we downsample the points?
                if (k % downsample == 0)
                  {
                    const vector3 point = calib.camera_to_world (vector3 (x, y, z));
                
                    if (!use_depth_threshold || (point [2] > depth_threshold && point [1] > y_thresh)) 
                      {
                        double d = -1;
                        if (!use_background_subtraction || (d = bg_nearest_cloud.nearest_distance (point)) > 0.5)
                          {
                            points.push_back (point);
                            colours.push_back (vector3 (r, g, b));
                          }
                      }
                  }
              }
            fclose (fid);
            // Render background
            //render_background_image (background, PROJ2D_SCALE);
            //proj2d.update_segment (background);
            
            // Update kNN structure
            nearest_cloud.refresh (points);
          }

        return found;
      }

    /**
     Return the surface normal of one 3D point in the background.
     This operation is currently not supported in this class.
     
     @param[in] i
       The index of the point you want to read.
       
     @return
       The 3D surface normal at the given index or a zero-vector if the index is too large.
       In this class a zero-vector is always returned.
     */
    vector3 bg_normal (const size_t i) const
      {
        return vector3 (0, 0, 0);
      }

  protected:
    const std::string input_template;
};

#endif // POINT_CLOUD_H
