// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#include <OpenTissue/configuration.h>

#define DEFINE_GLUT_MAIN
#include <OpenTissue/utility/glut/glut_perspective_view_application.h>
#undef DEFINE_GLUT_MAIN

#include "skeleton_cloud_measure.h"
#include "pf_skeleton_tracker.h"
#include "auxil.h"
#include "options.h"
#include "macros.h"

/**
 A simple class for posing a skeleton according to data.
 */
template <class observation_type>
class Application : public OpenTissue::glut::PerspectiveViewApplication
{
protected:
  calibration calib;
  angle_state<observation_type> pose;
  observation_type observation;
  int theta_idx;
  const double delta_theta;
  vector3 trans, rot;
  measure<angle_state<observation_type>, observation_type> meas;
  mocap_data<observation_type> mocap;
  bool in_progress;

public:
  Application (const options &opts)
    : calib (opts),
      pose (opts, calib),
      observation (opts, calib),
      theta_idx (0),
      delta_theta (0.1),
      meas (opts, calib),
      mocap (opts, calib),
      in_progress (false)
    {
      this->z_far () = 5000; // what does this do?
      trans = calib.get_world_translation ();
      rot = calib.get_world_rotation ();
    }

  const char* do_get_title () const
    {
      return "Poser";
    }

  void do_display ()
    {
      static bool showing = false;
      if (showing || in_progress) {std::cerr << "%%%%" << std::endl; return; }
      showing = true;
      
      glPushMatrix ();
//      glTranslatef (0, -6, -4);
      glScalef (0.5, 0.5, 0.5);
      pose.show ();
      observation.show ();
      
      OpenTissue::gl::ColorPicker (0.1, 0.1, 0.9, 1.0);
      OpenTissue::gl::DrawPoint (vector3 (0, 0, 0));

      mocap.show ();
      glPopMatrix ();
      showing = false;
    }
  
  void do_action (unsigned char choice)
    {
      in_progress = true;
      static bool do_trans = true;
      bool dirty = false;
      const double delta_trans = 0.1, delta_rot = 0.1 /*0.01*/;
      double tmp;
      double min_limit, max_limit;
      switch (choice)
        {
          case 'w':
            theta_idx = std::max (--theta_idx, 0);
            break;
          case 'e':
            theta_idx = std::min (++theta_idx, pose.theta_size ()-1);
            break;
          case 's':
            tmp = pose.get_theta (theta_idx) - delta_theta;
            min_limit = pose.get_min_limit (theta_idx);
            pose.set_theta (theta_idx, std::max (tmp, min_limit));
            pose.set_theta (theta_idx, tmp);
            dirty = true;
            break;
          case 'd':
            tmp = pose.get_theta (theta_idx) + delta_theta;
            max_limit = pose.get_max_limit (theta_idx);
            pose.set_theta (theta_idx, std::min (tmp, max_limit));
            pose.set_theta (theta_idx, tmp);
            dirty = true;
            break;
          case 'x':
            observation.refresh (false);
            mocap.refresh (false);
            break;
          case 'c':
            std::cout << "Saving to 'saved.pose'" << std::endl;
            pose.save ("saved.pose");
            break;
          case 'u':
            do_trans = !do_trans;
            if (do_trans)
              std::cout << "Translating" << std::endl;
            else
              std::cout << "Rotating" << std::endl;
            break;
          case 'g':
            if (do_trans)
              trans [0] += delta_trans;
            else
              rot [0] += delta_rot;
            calib.set_world_translation (trans);
            calib.set_world_rotation (rot);
            break;
          case 'h':
            if (do_trans)
              trans [1] += delta_trans;
            else
              rot [1] += delta_rot;
            calib.set_world_translation (trans);
            calib.set_world_rotation (rot);
            break;
          case 'j':
            if (do_trans)
              trans [2] += delta_trans;
            else
              rot [2] += delta_rot;
            calib.set_world_translation (trans);
            calib.set_world_rotation (rot);
            break;
          case 'b':
            if (do_trans)
              trans [0] -= delta_trans;
            else
              rot [0] -= delta_rot;
            calib.set_world_translation (trans);
            calib.set_world_rotation (rot);
            break;
          case 'n':
            if (do_trans)
              trans [1] -= delta_trans;
            else
              rot [1] -= delta_rot;
            calib.set_world_translation (trans);
            calib.set_world_rotation (rot);
            break;
          case 'm':
            if (do_trans)
              trans [2] -= delta_trans;
            else
              rot [2] -= delta_rot;
            calib.set_world_translation (trans);
            calib.set_world_rotation (rot);
            break;
          case 't':
            if (calib.save_world ("trans_and_rot.view"))
              std::cout << "Saved translation and rotation to 'trans_and_rot.view'" << std::endl;
            else
              std::cerr << "Could not save translation and rotation" << std::endl;
            break;
          default:
            std::cout << "You pressed " << choice << std::endl;
            break;
        };
      
      OpenTissue::kinematics::inverse::compute_joint_limits_projection (pose.get_skeleton (), pose.get_theta ());
      if (dirty)
        {
          pose.compute_pose ();
          pose.prepare_mesh ();
        }
      const double score = mocap.average_pose_dist (pose);
      const double m = 0; //meas.log_measure (pose, observation);
      std::cout << theta_idx << " " << pose.get_theta (theta_idx) << " "
                << m << " " << exp (m) << " "
                << score << " "
                << std::endl;
      
      in_progress = false;
    }

  void do_init_right_click_menu (int main_menu, void menu (int entry))
    {
    }

  void do_init ()
    {
      this->camera ().move (80);
      glClearColor (0.9, 0.9, 0.9, 0.0);
      std::cout << "Press 't' to get going" << std::endl;
    }

  void do_run ()
    {
    }

  void do_shutdown ()
    {
    }
};

#define EXPAND(STATE_NAME, STATE_TYPE, OBS_NAME, OBS_TYPE)                     \
        if (st == STATE_NAME && ot == OBS_NAME)                                \
          {                                                                    \
            instance.reset (new Application< OBS_TYPE > (opts));               \
          }

OpenTissue::glut::instance_pointer init_glut_application (int argc, char *argv [])
{
  // Disable localisation as that screws with punctuation in numerical IO
  setenv ("LANG", "", 1);

  // Get options
  const std::string options_filename = (argc > 1) ? argv [1] : "";
  const options opts (options_filename);

  // Create GUI
  const tracker_state_type st = opts.state_type ();
  const tracker_observation_type ot = opts.observation_type ();

  OpenTissue::glut::instance_pointer instance;
  
  EXPAND_ALL_TRACKERS
        
  return instance;
}

