// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef PROJECT2D
#define PROJECT2D

#include "opencv_helper.h"
#include "auxil.h"

//#define PROJ2D_SCALE 0.25
#define PROJ2D_SCALE 1.0

/**
 Project image points on a segment in the image. This is useful for finding the
 nearest point on a segment in an image. The algorithm computes the distance map
 of the segmented image and uses this to find the nearest point on a segment by
 calculating derivatives of the map. This operation can be unstable near ridges
 of the distance map.
 */
class project2d
{
public:
  /**
   Initialise the object. This function does not allocate any data.
   
   Prior to performing any projection you need to call the update_segment
   function at least once.
   */
  project2d ()
    : dist (NULL)
    {
      size = cvSize (0, 0);
    }
  
  /**
   Compute a new distance map for the projection operation. You only need to call
   this function when the segmented image changes.
   
   @param[in] new_segment
     A binary (8-bit gray scale) image containing the segment. The binary image
     is inverted as part of the computations.
   */
  void update_segment (IplImage *new_segment)
    {
      // Should we resize the distance map?
      const CvSize new_size = cvGetSize (new_segment);
      if (size.width != new_size.width || size.height != new_size.height)
        {
          if (dist != NULL)
            cvReleaseImage (&dist);
          dist = cvCreateImage (new_size, IPL_DEPTH_32F, 1);
          size = new_size;
        }
      
      // Compute distance map
      cvNot (new_segment, new_segment);
      cvDistTransform (new_segment, dist);
    }
  
  /**
   Project an image point onto the nearest segment in the binary image.
   
   @param[in] p
     Coordinates of an image point.
   
   @return
     The coordinates of the nearest image point on the segment. Due to the approximate
     nature of the algorithm, this point might not be on the segment, but it will
     be near.
   */
  CvPoint project (CvPoint p) const
    {
      CvPoint retval;

      // Ensure we are within the image before accessing pixels
      p.x = force_in_range (p.x, 0, size.width-1);
      p.y = force_in_range (p.y, 0, size.height-1);
  
      // Calculate derivative of 'dist' at 'p'
      const float val = GRAY32F (dist, p.x, p.y);
      if (val > 0) // outside
        {
          const float dx = 0.5 * (GRAY32F (dist, p.x + 1, p.y) - GRAY32F (dist, p.x - 1, p.y));
          const float dy = 0.5 * (GRAY32F (dist, p.x, p.y + 1) - GRAY32F (dist, p.x, p.y - 1));
          const float norm = sqrt (dx*dx + dy*dy);
          retval.x = p.x - dx * val / norm;
          retval.y = p.y - dy * val / norm;
        }
      else // inside (no need for projection)
        {
          retval.x = p.x;
          retval.y = p.y;
        }

      return retval;
    }
  
  /**
   Determines if an image point is inside a segment in the binary image.
   
   @param[in] p
     Coordinates of an image point.
   
   @return
     true if inside; otherwise false.
   */
  bool is_inside (CvPoint p) const
    {
      // Ensure we are within the image before accessing pixels
      p.x = force_in_range (p.x, 0, size.width-1);
      p.y = force_in_range (p.y, 0, size.height-1);
  
      // Check status
      const float val = GRAY32F (dist, p.x, p.y);
      return (val <= 0.0);
    }
  
  /**
   Return the size of the binary image from which the segment was found.
   
   @return
     The size of the image.
   */
  CvSize get_size () const
    {
      return size;
    }

private:
  IplImage *dist;
  CvSize size;
};

#endif // PROJECT2D
