// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef PROJECTION_STATE_H
#define PROJECTION_STATE_H

#include "angle_state.h"
#include "options.h"
#include "simulator.h"

/**
 An implementation of the projected prior. If you use this class we recommend
 citing
 @code
 @inproceedings{spatial_priors:hauberg_et_al10,
  title = {{Gaussian-like Spatial Priors for Articulated Tracking}},
  author = {S{\o}ren Hauberg and Stefan Sommer and Kim Steenstrup Pedersen},
  booktitle = {ECCV},
  year = "2010",
  pages = "425--437",
  editor = {K. Daniilidis and P. Maragos and N. Paragios},
  publisher = {Springer},
  series = {LNCS},
  volume = {6311}
 }
 @endcode
 */
template<class observation_type>
class projection_state : public skeleton_state<observation_type>
{
public:
  projection_state (const options &opts, const calibration &_calib, const bool _for_show = true)
    : skeleton_state<observation_type> (opts, _calib, _for_show),
      pred_noise (opts.pred_noise ())
    {
      this->compute_pose ();
  
      // Store default chains temporarely
      std::list<vector3> chain_list;
      for (chain_iter iter = this->solver.chain_begin (); iter != this->solver.chain_end (); iter++)
        chain_list.push_back (iter->get_end_effector ()->absolute ().T ());
  
      // Add everything as end-effectors
      int i = 0;
      for (bone_iter iter = this->bone_begin (); iter != this->bone_end (); iter++, i++)
        {
          if (iter->is_root ()) // || i == 3 || i == 4 || i == 5 ||  i == 7 || i == 11)
            continue;
        
          const vector3 absolute = iter->absolute ().T ();
          const vector3 pabsolute = iter->parent ()->absolute ().T ();

          // Make sure the bone isn't already an end-effector
          bool accept = true; //(absolute != pabsolute);
          for (std::list<vector3>::const_iterator ci = chain_list.begin ();
               ci != chain_list.end (); ci ++)
            {
              if (*ci == absolute)
                {
                  accept = false;
                  break;
                }
            }
      
          if (accept)
            {
              chain_type chain;
              chain.init (this->skeleton.root (), &(*iter));
              this->solver.add_chain (chain);
            }
        }
    };

  double predict (const observation_type &observation, const double variance_scale = 1.0)
    {
      gaussian1D gauss;

      // Predict root
      //this->root (0) += 0.05 * gauss.random ();
      //this->root (1) += 0.0001 * gauss.random ();
      //this->root (2) += 0.05 * gauss.random ();

      this->compute_pose ();
      vector3 nu;
      int k = 0;

      const double std [] = {0.019905, 0.003157, 0.008050, 0.180706, 0.101031, 0.137931,
                             0.275119, 0.182599, 0.286416, 0.275119, 0.182599, 0.286416,
                             0.390496, 0.301062, 0.483356, 0.508283, 0.808703, 0.806479,
                             0.275119, 0.182599, 0.286416, 0.381652, 0.255074, 0.218130,
                             0.498532, 0.249264, 0.615678, 0.938598, 0.593349, 1.098928,
                             0.275119, 0.182599, 0.286416, 0.509799, 0.464794, 0.370578,
                             0.988121, 0.796506, 0.450776, 1.155469, 1.320700, 0.709060};
  
      for (chain_iter iter = this->solver.chain_begin (); iter != this->solver.chain_end (); iter++)
        {
          for (size_t n = 0; n < 3; n++)
            nu [n] = 0.5 * std [k++] * gauss.random () / variance_scale;

          iter->p_global () = iter->get_end_effector ()->absolute ().T () + nu; // - this->root;
        }

      this->solver.solve (&solver_type::default_SD_settings ());
      OpenTissue::kinematics::inverse::get_joint_parameters (*(this->solver.skeleton ()),
                                                             this->theta);

      return 0;
    };
  
  projection_state& operator= (projection_state &new_state)
    {
      skeleton_state<observation_type>::operator= (new_state);

      return *this;
    };
    
  bool load (const std::string filename)
    {
      return skeleton_state<observation_type>::load (filename);
    };

  bool load (std::ifstream &file)
    {
      return skeleton_state<observation_type>::load (file);
    };

  private:
    const double pred_noise;
};

#endif // PROJECTION_STATE_H
