// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#include <cmath>
#include <iostream>
#include <cassert>
#include "simulator.h"
#include <ctime>
#include <stdlib.h>
#include <math.h>

#include "auxil.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

double uniform::random () const
{
  return rand () / (double)RAND_MAX;
}

double uniform::pdf (const double y) const
{
  if (y < 0 || y > 1)
    return 0.0;
  else
    return 1.0;
}

double uniform::cdf (const double y) const
{
  if (y < 0)
    {
      return 0.0;
    }
  else if (y >= 0 && y<=1)
    {
      return y;
    }
  else
    {
      return 1.0;
    }
}

gaussian1D::gaussian1D (const double m, const double s)
  : mu (m),
    sigma (s)
{
}

double gaussian1D::random () const
{
  const uniform un;
  return mu + sigma * sqrt (-2.0 * log (un.random ())) * cos (2.0 * M_PI * un.random ());
}

double gaussian1D::pdf (const double y) const
{
  return exp (-square (y - mu) / (2.0 * square (sigma)))/(sqrt (2.0 * M_PI) * sigma);
}

double gaussian1D::cdf (const double y) const
{
  std::cerr << "no explicite cdf for gaussian" << std::endl;
  exit (1);
}

von_mises::von_mises (const double m, const double k)
  : mu (m), kappa (k)
{
}

double von_mises::pdf (const double x) const
{
  const double p = exp (kappa * cos (x - mu));
  const double Z = 2.0 * M_PI * bessi0 (kappa);
  return p / Z;
}

double von_mises::random () const
{
  double theta;
  if (kappa < 1e-6)
    {
      // kappa is small: sample uniformly on circle
      theta = 2.0 * M_PI * uni.random () - M_PI;
    }
  else
    {
      const double a = 1.0 + sqrt (1.0 + 4.0 * kappa * kappa);
      const double b = (a - sqrt (2.0 * a)) / (2.0 * kappa);
      const double r = (1.0 + b*b) / (2.0 * b);

      bool not_done = true;
      double f;
      while (not_done)
        {
          const double u1 = uni.random ();
          const double u2 = uni.random ();
      
          const double z = cos (M_PI * u1);
          f = (1.0 + r * z) / (r + z);
          const double c = kappa * (r - f);
      
          not_done = (u2 >= c * (2.0 - c)) && (log (c) - log (u2) + 1.0 - c < 0.0);
        }
    
      const double u3 = uni.random ();
      theta = mu + sign (u3 - 0.5) * acos (f);
    }
  
  return theta;
}

double von_mises::cdf (const double) const
{
  std::cerr << "von_mises::cdf not implemented" << std::endl;
  return -1;
}

exponential::exponential (const double m)
  : mu (m)
{
}

double exponential::pdf (const double y) const
{
  if (y < 0)
    return 0;
  else
    return mu * exp (-mu * y);
}

double exponential::cdf (const double y) const
{
  if (y < 0)
    return 0;
  else
    return 1.0 - exp (-mu * y);
}

double exponential::random () const
{
  const uniform  un;
  return -1.0 * log (1.0 - un.random ()) / mu;
}

chi_square::chi_square (const int n0)
  : n (n0)
{
}

double chi_square::random () const
{
  double chisquare = 0.0;
  const gaussian1D un;
  for (int k = 0; k < n; k++)
    {
      const double x = un.random ();
      chisquare += square (x);
    }
    
  return chisquare;
}

inline
double factorial (const int n0)
{
  int f = 1;
  for (int k = 1; k <= n0; k++)
    f *= k;
  
  return f;
}


static
double gamma2 (const double t)
{
  // XXX: Why on earth do we have this check?? chi_sqaure::pdf will never call the function like this!
  if (2*(int)t != 2.0*t)
    {
      std::cerr << "error" << std::endl; // XXX: What kind of error occured?
      exit (1);
    }

  if (t == 0.5)
    return sqrt (M_PI);
  else
    return (t-1.0) * gamma2 (t); // XXX: infinite loop!!!
}

double chi_square::pdf (const double y) const
{
  const double gamma = (n%2 == 0) ? factorial (n/2) : gamma2 (n/2.0);
  return pow (y, 0.5*n) * exp (-0.5*y) / (gamma * pow (2.0, 0.5*n));
}

double chi_square::cdf (const double y) const
{
  std::cerr << "not implemented" << std::endl;
  exit (1);
}

rayleigh::rayleigh (const double m)
  : mu (m)
{
}

double rayleigh::pdf (const double y) const
{
  if (y <= 0)
    return 0.0;
  else
    return M_PI * y * exp (-M_PI * square (y) / (4.0 * square (mu))) / (2.0 * square (mu));
}

double rayleigh::cdf (const double y) const
{
  if (y <= 0)
    return 0.0;
  else
    return 1.0 - exp (-M_PI * square (y) / (4.0 * square (mu)));
}

double rayleigh::random () const
{
  const uniform  un;
  return sqrt (-4.0 * square (mu) * log (1.0 - un.random ()) / M_PI);
}

discrete::discrete (const std::vector<double> &p)
  : prob (p),
    cum_sum (p.size ()),
    nb (p.size ())
{
  cum_sum [0] = prob [0];
  for (int k = 1; k < nb; k++)
    cum_sum [k] = p [k] + cum_sum [k-1];
}

int discrete::random () const
{
  int discr = nb - 1;
  
  const uniform ud;
  const double r = ud.random ();

  // XXX: Use binary search instead (cum_sum is sorted)
  for (int k = 0; k < nb; k++)
    if (r <= cum_sum [k])
      {
        discr = k;
        break; 
      }

  return discr;
}

double discrete::pdf (const int k) const
{
  return prob [k];
}

double discrete::cdf (const int k) const
{
  return cum_sum [k];
}

