// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef SIMULATOR
#define SIMULATOR

#include <vector>

/**
 A base class for random nuber generators.
 */
template<typename T> class simulator
{
public:
  virtual T random () const = 0; // simulate
  virtual double pdf (const T) const = 0; // probability density
  virtual double cdf (const T) const = 0; // cumulative density
  virtual ~simulator ()
    {
    }
};

/**
 A uniform random number generator.
 */
class uniform : public simulator<double>
{
public:
  double pdf (const double) const;
  double cdf (const double) const;
  double random () const;
};

/**
 A Gaussian random number generator.
 */
class gaussian1D : public simulator<double>
{
public:
  gaussian1D (const double m = 0, const double s = 1);
  double pdf (const double) const;
  double random () const;
  double cdf (const double) const;

private:
  const double mu;
  const double sigma;
};

/**
 A Von-Mises random number generator.
 */
class von_mises : public simulator<double>
{
public:
  von_mises (const double m = 0, const double k = 1);
  double pdf (const double) const;
  double random () const;
  double cdf (const double) const;

private:
  const double mu;
  const double kappa;
  uniform uni;
};

/**
 An Exponential random number generator.
 */
class exponential : public simulator<double>
{
public:
  exponential (const double m = 1.0);
  double pdf (const double) const;
  double random () const;
  double cdf (const double) const;

private:
  const double mu;
};

/**
 A Chi-Square random number generator.
 */
class chi_square : public simulator<double>
{
public:
  chi_square (const int n0 = 1);
  double cdf (const double) const;
  double pdf (const double) const;
  double random () const;

private:
  const int n;
};

/**
 A Rayleigh random number generator.
 */
class rayleigh : public simulator<double>
{
public:
  rayleigh (const double);
  double pdf (const double) const;
  double cdf (const double) const;
  double random () const;

private:
  const double mu;
};

/**
 A discrete random number generator.
 */
class discrete : public simulator<int>
{
public:
  discrete (const std::vector<double> &p);
  double pdf (const int) const;
  double cdf (const int) const;
  int random () const;

private:
  const std::vector<double> prob;
  std::vector<double> cum_sum;
  const int nb;
};

#endif
