// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#include <OpenTissue/configuration.h>

#define DEFINE_GLUT_MAIN
#include <OpenTissue/utility/glut/glut_perspective_view_application.h>
#undef DEFINE_GLUT_MAIN

#include "pf_skeleton_tracker.h"
#include "auxil.h"
#include "options.h"
#include "skeleton_cloud_measure.h"
#include "macros.h"

template <class observation_type>
bone_iter get_bone (angle_state<observation_type> &pose, int idx)
{
  int k = 0;
  bone_iter retval = pose.bone_begin ();
  while (k < idx && retval != pose.bone_end ())
    {
      k++;
      retval++;
    }
  return retval;
}

template <class observation_type>
void dump_result (angle_state<observation_type> &pose)
{
  for (bone_iter iter = pose.bone_begin (); iter != pose.bone_end (); iter++)
    {
      std::cout << iter->get_name () << " " << iter->relative ().T () << std::endl;
    }
}

/**
 A simple application for changin lengths of bones in the skeleton.
 */
template <class observation_type>
class Application : public OpenTissue::glut::PerspectiveViewApplication
{
protected:
  calibration calib;
  angle_state<observation_type> pose;
  observation_type observation;
  int bone_idx;
  const double delta_theta;
  measure<angle_state<observation_type>, observation_type> meas;

public:
  Application (const options &opts)
    : calib (opts),
      pose (opts, calib),
      observation (opts, calib),
      bone_idx (0),
      delta_theta (0.1),
      meas (opts, calib)
    {
      this->z_far () = 5000; // what does this do?
    }

  const char* do_get_title () const
    {
      return "Skeleton Sizer";
    }

  void do_display ()
    {
      static bool in_progress = false;
      if (in_progress) return;
      
      in_progress = true;
      glPushMatrix ();
      glTranslatef (0, -3, 0);
      pose.show ();
      observation.show ();
      glPopMatrix ();
      in_progress = false;
    }

  void do_action (unsigned char choice)
    {
      bone_iter bone = get_bone (pose, bone_idx);
      const double delta = 0.05;
      switch (choice)
        {
          case 'w':
            bone_idx = std::max (--bone_idx, 0);
            bone = get_bone (pose, bone_idx);
            break;
          case 'e':
            bone_idx = std::min (++bone_idx, pose.theta_size ()); // XXX: use a different maximum
            bone = get_bone (pose, bone_idx);
            break;
          case 's':
            bone->relative ().T () *= (1.0 - delta);
            break;
          case 'd':
            bone->relative ().T () *= (1.0 + delta);
            break;
          case 'x':
            observation.refresh ();
            break;
          case 'c':
            dump_result (pose);
            std::cout << "XXX: Saving to 'saved.pose'" << std::endl;
            break;
          default:
            std::cout << "You pressed " << choice << std::endl;
            break;
        };
      
      std::cout << bone_idx << " " << bone->get_name () << " " << std::endl;
    }

  void do_init_right_click_menu (int main_menu, void menu (int entry))
    {
    }

  void do_init ()
    {
      this->camera ().move (80);
      glClearColor (0.9, 0.9, 0.9, 0.0);
      std::cout << "Press 't' to get going" << std::endl;

      /*
      for (bone_iter iter = pose.bone_begin (); iter != pose.bone_end (); iter++)
        {
          iter->relative ().T () *= 1.25;
        }
      */
    }

  void do_run ()
    {
    }

  void do_shutdown ()
    {
    }
};

#define EXPAND(STATE_NAME, STATE_TYPE, OBS_NAME, OBS_TYPE)                     \
        if (st == STATE_NAME && ot == OBS_NAME)                                \
          {                                                                    \
            instance.reset (new Application< OBS_TYPE > (opts));               \
          }

OpenTissue::glut::instance_pointer init_glut_application (int argc, char *argv [])
{
  // Disable localisation as that screws with punctuation in numerical IO
  setenv ("LANG", "", 1);

  const std::string options_filename = (argc > 1) ? argv [1] : "";
  const options opts (options_filename);

  // Create GUI
  const tracker_state_type st = opts.state_type ();
  const tracker_observation_type ot = opts.observation_type ();

  OpenTissue::glut::instance_pointer instance;
  
  EXPAND_ALL_TRACKERS

  return instance;
}

