// Copyright (C) 2011 Soren Hauberg
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>.
 
#ifndef STATE_H
#define STATE_H

#include <string>
#include <fstream>

/**
 An abstract class for representing the state of the particle filter.
 */
template <class observation_type>
class state
{
  public:
    virtual ~state () {}; // Just here to remove a warning
    /**
     The importance distribution of the tracker.
     
     @param [in] observation
     The current observation.
     
     @param [in] variance_scale
     A scale factor for the default variance. This is used for annealed filters.
     
     @return
     The correction factor for the importance distribution (most often 1).
     */
    virtual double predict (const observation_type &observation, const double variance_scale = 1.0) = 0;
    
    /**
     Visualise the state.
     
     @param [in] gray_level
     The gray-scale value when drawing the state.
     */
    virtual void show (const double gray_level = 0.5) const = 0;
    
    /**
     Save the state to a file.
     
     @param [in] filename
     The output filename.
     */
    virtual void save (const std::string filename) const = 0;
    
    /**
     Save the state to a stream.
     
     @param [in] file
     The output stream.
     */
    virtual void save (std::ofstream &file) const = 0;
    
    /**
     Write relevant state information to the log.
     
     @param [in] file
     The stream of the log file.
     */
    void write_to_log (std::ofstream &file) const
      {
        // This function is optional, so individual state classes should implement
        // the needed behaviour
      };
};

#endif // STATE_H

